"""Subscription class: for MQSUB, MQSUBRQ, MQCLOSE.
Will reference a managed queue for the MQGET if that has been requested.
"""

# Copyright (c) 2025, 2026 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from typing import Union

from mqcommon import *
from mqerrors import *
from ibmmq import CMQC, SD, SRO, MQObject, Queue, ibmmqc
import mqlog

class Subscription(MQObject):
    """ Encapsulates a subscription to a topic.
    """
    def __init__(self, queue_manager, sub_desc=None, sub_name=None,
                 sub_queue=None, sub_opts=None, topic_name=None, topic_string=None):

        mqlog.trace_entry("sub:__init__")

        queue_manager = ensure_strings_are_bytes(queue_manager)
        sub_name = ensure_strings_are_bytes(sub_name)
        topic_name = ensure_strings_are_bytes(topic_name)
        topic_string = ensure_strings_are_bytes(topic_string)

        self.__queue_manager = queue_manager
        self.sub_queue = sub_queue
        self.__sub_desc = sub_desc
        self.sub_name = sub_name
        self.sub_opts = sub_opts
        self.topic_name = topic_name
        self.topic_string = topic_string
        self.__sub_handle = None

        if self.__sub_desc:
            self.sub(sub_desc=self.__sub_desc)

        object_name = None
        if topic_name:
            object_name = topic_name
        if topic_string:
            if object_name:
                object_name = object_name + "/" + topic_string
            else:
                object_name = topic_string
        super().__init__(object_name)
        mqlog.trace_exit("sub:__init__")

    def get_handle(self):
        """Return the subscription's hObj object"""
        return self.__sub_handle

    def get_queue_manager(self):
        """Return the subscriptions hConn object"""
        return self.__queue_manager

    def get_sub_queue(self) -> Queue:
        """ Return the subscription queue.
        """
        return self.sub_queue

    def get(self, max_length: Union[None, int] = None, *opts):  # pylint: disable=keyword-arg-before-vararg
        """ Get a publication from the Queue.
        """
        mqlog.trace_entry("sub:get")
        rv = self.sub_queue.get(max_length, *opts)
        mqlog.trace_exit("sub:get")

        return rv

    def get_rfh2(self, max_length: Union[None, int] = None, *opts) -> bytes:  # pylint: disable=keyword-arg-before-vararg
        """ Get a publication from the Queue.
        """
        mqlog.trace_entry("sub:get_rfh2")
        rv = self.sub_queue.get_rfh2(max_length, *opts)
        mqlog.trace_exit("sub:get_rfh2")
        return rv

    def sub(self, sub_desc=None, sub_queue=None, sub_name=None, sub_opts=None,
            topic_name=None, topic_string=None):
        """ Subscribe to a topic, alter or resume a subscription.
        Executes the MQSUB call with parameters.
        The subscription queue can be either passed as a Queue object or a
        Queue object handle.
        """
        mqlog.trace_entry("sub:sub")

        sub_queue = ensure_strings_are_bytes(sub_queue)
        sub_name = ensure_strings_are_bytes(sub_name)
        topic_name = ensure_strings_are_bytes(topic_name)
        topic_string = ensure_strings_are_bytes(topic_string)

        if topic_name:
            self.topic_name = topic_name
        if topic_string:
            self.topic_string = topic_string
        if sub_name:
            self.sub_name = sub_name

        if sub_desc:
            if not isinstance(sub_desc, SD):
                mqlog.trace_exit("sub:sub", ep=1)
                raise TypeError('sub_desc must be a SD(sub descriptor) object.')
        else:
            sub_desc = SD()
            if sub_opts:
                sub_desc['Options'] = sub_opts
            else:
                sub_desc['Options'] = CMQC.MQSO_CREATE + CMQC.MQSO_NON_DURABLE + CMQC.MQSO_MANAGED
            if self.sub_name:
                sub_desc.set_vs('SubName', self.sub_name)
            if self.topic_name:
                sub_desc['ObjectName'] = self.topic_name
            if self.topic_string:
                sub_desc.set_vs('ObjectString', self.topic_string)
        self.__sub_desc = sub_desc

        sub_queue_handle = CMQC.MQHO_NONE
        if sub_queue:
            if isinstance(sub_queue, Queue):
                sub_queue_handle = sub_queue.get_handle()
            else:
                sub_queue_handle = sub_queue

        rv = ibmmqc.MQSUB(self.__queue_manager.getHandle(), sub_desc.pack(), sub_queue_handle)

        if rv[-2]:
            mqlog.trace_exit("sub:sub", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        sub_desc.unpack(rv[0])
        self.__sub_desc = sub_desc
        self.sub_queue = Queue(self.__queue_manager)
        self.sub_queue.set_handle(rv[1])
        self.__sub_handle = rv[2]

        if OTelFunctions.open and (sub_desc.Options & CMQC.MQSO_MANAGED) != 0:
            OTelFunctions.open(self, None, 0, self.sub_queue)
        mqlog.trace_exit("sub:sub")

    def close(self, sub_close_options=CMQC.MQCO_NONE, close_sub_queue=False, close_sub_queue_options=CMQC.MQCO_NONE):
        """Close the subscription"""
        mqlog.trace_entry("sub:close")

        if not self.__sub_handle:
            mqlog.trace_exit("sub:close", ep=1)
            raise PYIFError('Subscription not open.')

        rv = ibmmqc.MQCLOSE(self.__queue_manager.getHandle(), self.__sub_handle, sub_close_options)
        if rv[0]:
            mqlog.trace_exit("sub:close", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        if OTelFunctions.close:
            OTelFunctions.close(self)

        self.__sub_handle = None
        self.__sub_desc = None

        if close_sub_queue:
            _ = self.sub_queue.close(close_sub_queue_options)

        mqlog.trace_exit("sub:close")

    def subrq(self, sub_action: int = CMQC.MQSR_ACTION_PUBLICATION, sro=None) -> None:
        """Call the MQSUBRQ function. If the SRO object is supplied then it
        may be updated by the operation.
        """
        mqlog.trace_entry("sub:subrq")

        if sro:
            if not isinstance(sro, SRO):
                mqlog.trace_exit("sub:subrq", ep=1)
                raise TypeError('sro must be an SRO(sub request options) object.')
        else:
            sro = SRO()

        rv = ibmmqc.MQSUBRQ(self.__queue_manager.getHandle(), self.__sub_handle, sub_action, sro.pack())

        if rv[-2]:
            mqlog.trace_exit("sub:subrq", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])

        _ = sro.unpack(rv[0])
        mqlog.trace_exit("sub:subrq")

    def __del__(self):
        """ Close the Subscription, if it has been opened.
        """
        mqlog.trace_entry("sub:__del__")

        try:
            if self.__sub_handle:
                self.close()
        except PYIFError:
            pass

        mqlog.trace_exit("sub:__del__")



"""MQCMHO, MQDMHO, MQPD, MQIMPO, MQSMPO, MQDMPO: Structures for
controlling Message Handles and Properties
"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class CMHO(MQOpts):
    """ Construct an MQCMHO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQCMHO_STRUC_ID, '4s'],
                ['Version', CMQC.MQCMHO_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQCMHO_DEFAULT_VALIDATION, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class DMHO(MQOpts):
    """ Construct an MQDMHO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQDMHO_STRUC_ID, '4s'],
                ['Version', CMQC.MQDMHO_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQDMHO_NONE, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class PD(MQOpts):
    """ Construct an MQPD Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQPD_STRUC_ID, '4s'],
                ['Version', CMQC.MQPD_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQPD_NONE, MQLONG_TYPE],
                ['Support', CMQC.MQPD_SUPPORT_OPTIONAL, MQLONG_TYPE],
                ['Context', CMQC.MQPD_NO_CONTEXT, MQLONG_TYPE],
                ['CopyOptions', CMQC.MQCOPY_DEFAULT, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class SMPO(MQOpts):
    """ Construct an MQSMPO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQSMPO_STRUC_ID, '4s'],
                ['Version', CMQC.MQSMPO_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQSMPO_SET_FIRST, MQLONG_TYPE],
                ['ValueEncoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
                ['ValueCCSID', CMQC.MQCCSI_APPL, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class DMPO(MQOpts):
    """ Construct an MQDMPO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQDMPO_STRUC_ID, '4s'],
                ['Version', CMQC.MQDMPO_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQDMPO_DEL_FIRST, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class IMPO(MQOpts):
    """ Construct an MQIMPO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQIMPO_STRUC_ID, '4s'],
                ['Version', CMQC.MQIMPO_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQIMPO_INQ_FIRST, MQLONG_TYPE],
                ['RequestedEncoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
                ['RequestedCCSID', CMQC.MQCCSI_APPL, MQLONG_TYPE],
                ['ReturnedEncoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
                ['ReturnedCCSID', (0), MQLONG_TYPE],
                ['_Reserved1', (0), MQLONG_TYPE],

                # ReturnedName
                ['ReturnedNameVSPtr', 0, 'P'],
                ['ReturnedNameVSOffset', (0), MQLONG_TYPE],
                ['ReturnedNameVSBufSize', (0), MQLONG_TYPE],
                ['ReturnedNameVSLength', (0), MQLONG_TYPE],
                ['ReturnedNameVSCCSID', (0), MQLONG_TYPE],

                ['TypeString', b'', '8s']]

        super().__init__(tuple(opts), **kw)



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example browses a queue, formatting any known MQ header structures that it finds. It is
an extension to the DLH example, in that it deals with a number of other headers
including the XQH found on messages on transmission queues.
"""

import ibmmq as mq
from ibmmq import CMQC

queue_manager = 'QM1'
channel = 'DEV.ADMIN.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'admin'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

ok = True

od = mq.OD()
# This is an XMITQ associated with a STOPPED channel, because we want to look at XQH processing
od.ObjectName = 'QM2.STOPPED'

q = mq.Queue(qmgr, od, CMQC.MQOO_BROWSE)

gmo = mq.GMO()
gmo.Options = CMQC.MQGMO_BROWSE_FIRST
cnt = 1

while ok:
    try:
        md = mq.MD()
        msg = q.get(None, md, gmo)

        print('------------------------------------')
        print(f"Message: {cnt}")
        cnt += 1

        gmo.Options = CMQC.MQGMO_BROWSE_NEXT
        fmt = md['Format']

        headers = True
        offset = 0
        ccsid = md['CodedCharSetId']

        # Iterate through headers that we might expect to see until there are no more.
        while headers:
            print()
            print('------------')
            print(f'Header: {bytes.decode(fmt, "utf8")}')
            print('------------')

            if fmt == CMQC.MQFMT_XMIT_Q_HEADER:
                # The XQH definition in Python has only the first part of the XQH in C, with the
                # embedded MQMD excluded. But We can extract both parts explicitly with XQH methods.
                xqh = mq.XQH().get_header(msg)
                print(xqh.to_string())

                # The to_string method changes structure contents, so we first stash the
                # Format in its bytes version. That matches the CMQC definition as a bytes-string.
                emd = mq.XQH().get_embedded_md(msg)
                fmt = emd['Format']
                ccsid = emd['CodedCharSetId']
                # print(f'G = {emd['GroupId']}')

                print()
                offset += CMQC.MQXQH_CURRENT_LENGTH

                print(emd.to_string())

            elif fmt == CMQC.MQFMT_MD_EXTENSION:
                # This will only be seen when looking at transmission queues with the XQH block
                mde = mq.MDE().unpack(msg[offset:offset + CMQC.MQMDE_CURRENT_LENGTH])
                fmt = mde['Format']
                ccsid = mde['CodedCharSetId']
                offset += CMQC.MQMDE_CURRENT_LENGTH

                print(mde.to_string())

            elif fmt == CMQC.MQFMT_RF_HEADER_2:
                rfh2 = mq.RFH2()
                rfh2.unpack(msg[offset:], 'utf8')
                fmt = rfh2['Format']
                offset += rfh2['StrucLength']

                print(rfh2.to_string())
                # Extract the separate folders ...
                folders = rfh2.get_folders()
                for folder in folders:
                    print(f'  Folder {folder} = ', rfh2[folder])

            elif fmt == CMQC.MQFMT_DEAD_LETTER_HEADER:
                dlh = mq.DLH()
                dlh.unpack(msg[offset:offset + CMQC.MQDLH_CURRENT_LENGTH])
                fmt = dlh['Format']
                ccsid = dlh['CodedCharSetId']
                offset += CMQC.MQDLH_CURRENT_LENGTH

                print(dlh.to_string())

            else:
                headers = False

        # And now print the message body. Strings get converted, otherwise just print bytes
        print()
        print('Message Body:')

        if fmt == CMQC.MQFMT_STRING:
            cp = 'utf8'  # Assume a codepage, though we might want to use the ccsid to be more discriminating
            print(bytes.decode(msg[offset:], cp))
        else:
            print(msg[offset:])

        print()

    except mq.MQMIError as e:
        if e.reason == CMQC.MQRC_NO_MSG_AVAILABLE:
            print('No more messages.')
            ok = False
        else:
            raise
q.close()
qmgr.disconnect()



"""QueueManager class: Implements MQI functions that do not require an hObj"""

# Copyright (c) 2025, 2026 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

import mqlog

from mqcommon import *
from ibmmq import CMQC, CMQXC, CNO, CSP, CTLO, SCO, PMO, STS, MD, OD, CD, MQObject, ibmmqc
from mqerrors import *

import mqcallback
import mqinq
import mqqargs

try:
    from typing import Any, Optional, Dict, Union
except ImportError:
    pass

class QueueManager(MQObject):
    """ QueueManager encapsulates the connection to the Queue Manager. By
    default, the Queue Manager is implicitly connected. If required,
    the connection may be deferred until a call to connect().
    """
    def __init__(self, name: Optional[Union[str, bytes]] = '', disconnect_on_exit: bool = True,
                 bytes_encoding: str = EncodingDefault.bytes_encoding,
                 default_ccsid: int = EncodingDefault.ccsid) -> None:
        """ Connect to the Queue Manager 'name' (default value '').
        If 'name' is None, don't connect now, but defer the connection until connect() is called.
        Input 'bytes_encoding'  and 'default_ccsid' are the encodings that will be used in PCF, MQPUT and MQPUT1 calls
        using this MQ connection in case Unicode objects should be given on input.
        """
        mqlog.trace_entry("qmgr:__init__")

        name = ensure_strings_are_bytes(name)

        self.__handle = None
        self.__name = name
        self.__disconnect_on_exit = disconnect_on_exit
        self.__qmobj = None
        self.bytes_encoding = bytes_encoding
        self.default_ccsid = default_ccsid

        if name is not None:
            self.connect(name)
        super().__init__(name)
        mqlog.trace_exit("qmgr:__init__")

    def __del__(self) -> None:
        """ Disconnect from the queue Manager, if connected. Ignore any
        errors from the MQI as there's nothing that can be done about it anyway.
        """
        mqlog.trace_entry("qmgr:__del__")
        if self.__handle:
            if self.__qmobj:
                try:
                    ibmmqc.MQCLOSE(self.__handle, self.__qmobj, CMQC.MQCO_NONE)
                except (PYIFError, MQMIError):
                    pass

            if self.__disconnect_on_exit:
                try:
                    self.disconnect()
                except (PYIFError, MQMIError):
                    pass
        mqlog.trace_exit("qmgr:__del__")

    # This is the simplest form of MQCONN and allows for no options. Not even authentication.
    def connect(self, name) -> None:
        """connect(name)

        Connect immediately to the Queue Manager 'name'."""

        mqlog.trace_entry("qmgr:connect")
        rv = ibmmqc.MQCONN(name)
        if rv[1]:
            mqlog.trace_exit("qmgr:connect", ep=1, rc=rv[2])
            raise MQMIError(rv[1], rv[2])
        self.__handle = rv[0]
        self.__name = name
        mqlog.trace_exit("qmgr:connect")

# MQCONNX code courtesy of John OSullivan (mailto:jos@onebox.com)
# SSL additions courtesy of Brian Vicente (mailto:sailbv@netscape.net)
# Connect options suggested by Jaco Smuts (mailto:JSmuts@clover.co.za)

    def connect_with_options(self, name: Union[str, bytes], *args: Any, **kwargs: Dict[str, Any]) -> None:
        """connect_with_options(name
                                [,user=user][,password=password]
                                [,opts=cnoopts][,cno=mqcno][,sco=mqsco]
                                [,csp=mqcsp] [,bno=mqbno][,cd=mqcd])

           connect_with_options(name, cd, [sco])

        Connect immediately to the Queue Manager 'name', using the
        optional MQCNO, MQCD, MQSCO, MQCSP classes.  The cnoopts
        Options field is available for backwards compatibility, but
        setting it within an MQCNO class is preferred.

        The second form is defined for backward compatibility with
        very old versions of the library. It connects immediately to
        the Queue Manager 'name', using the MQCD connection descriptor
        cd and the optional MQSCO SSL options sco.

        UserId and Password can be given explicitly (for compatibility) or
        (preferred) as fields in the MQCSP structure where they are called
        CSPUserId and CSPPassword. Other authentication mechanisms - in particular
        using Tokens - require the CSP to be supplied.
        """
        mqlog.trace_entry("qmgr:connect_with_options")

        name = ensure_strings_are_bytes(name)

        # Deal with old style args
        len_args = len(args)
        if len_args:
            if len_args > 2:
                mqlog.trace_exit("qmgr:connect_with_options", ep=10)
                raise TypeError('Too many positional args provided')
            if len_args >= 1:
                kwargs['cd'] = args[0]
            if len_args == 2:
                kwargs['sco'] = args[1]

        restore_csp = False
        restore_cno = False
        restore_sco = False
        restore_cd = False

        csp_pack = None
        sco_pack = None
        cno_pack = None
        bno_pack = None
        cd_pack = None

        initial_key = None
        https_keystore = None
        key_repo_password = None
        ssl_peername = None

        ccdt_url = None

        options = kwargs['opts'] if 'opts' in kwargs else CMQC.MQCNO_NONE
        bno = kwargs['bno'] if 'bno' in kwargs else None

        cno = kwargs['cno'] if 'cno' in kwargs else None
        if cno:
            if not isinstance(cno, CNO):
                mqlog.trace_exit("qmgr:connect_with_options", ep=1)
                raise TypeError("cno must be an instance of CNO")

            # The only field we need to work on is the CCDTUrl
            try:
                ccdt_url = cno.CCDTUrl
                cno._set_ptr_field('CCDTUrl', ccdt_url)
            except AttributeError:
                ccdt_url = None

            restore_cno = True
        else:
            cno = CNO()
            cno.Options = options
        if cno:
            # Set the most useful handle sharing option automatically
            # unless there's already an explicit value
            if (cno.Options & (CMQC.MQCNO_HANDLE_SHARE_NO_BLOCK |
                               CMQC.MQCNO_HANDLE_SHARE_BLOCK)) == 0:
                cno.Options |= CMQC.MQCNO_HANDLE_SHARE_BLOCK

            cno_pack = cno.pack()

        sco = kwargs['sco'] if 'sco' in kwargs else None
        if sco:
            if not isinstance(sco, SCO):
                mqlog.trace_exit("qmgr:connect_with_options", ep=2)
                raise TypeError("sco must be an instance of SCO")

            try:
                key_repo_password = sco.KeyRepoPassword
                sco._set_ptr_field('KeyRepoPassword', key_repo_password)
            except AttributeError:
                key_repo_password = None

            try:
                https_keystore = sco.HTTPSKeyStore
                sco._set_ptr_field('HTTPSKeyStore', https_keystore)
            except AttributeError:
                https_keystore = None

            restore_sco = True

            sco_pack = sco.pack()

        csp = kwargs['csp'] if 'csp' in kwargs else None
        if csp:
            if not isinstance(csp, CSP):
                mqlog.trace_exit("qmgr:connect_with_options", ep=3)
                raise TypeError("csp must be an instance of CSP")
            # The real names of the user/password field also start with CSP
            # but it's very easy to forget that and we don't get compile errors
            # if people just set CSP.UserId. So throw an explicit error.
            # I considered silently permitting these "bad" fields and moving them to the
            # real names under the covers, but decided that it was
            # better to encourage use of the formal MQI structure elements. Even
            # though they're not really well-named. I don't do this for other MQI
            # fields, but this pair have proven to be particularly common mistakes.
            user = csp.CSPUserId
            try:
                _ = csp.UserId
                mqlog.trace_exit("qmgr:connect_with_options", ep=4)
                raise PYIFError('UserId field in CSP class is called CSPUserId')
            except AttributeError:
                pass

            password = csp.CSPPassword
            try:
                _ = csp.Password
                mqlog.trace_exit("qmgr:connect_with_options", ep=5)
                raise PYIFError('Password field in CSP class is called CSPPassword')
            except AttributeError:
                pass

            try:
                token = csp.Token
            except AttributeError:
                token = None
            try:
                initial_key = csp.InitialKey
            except AttributeError:
                initial_key = None

            # String fields in the CSP are overwritten with a ptr
            # so we will restore them to the original values after
            # the CONNX
            restore_csp = True
        else:
            # Always use a CSP even if none of the user fields are set
            csp = CSP()
            user = kwargs.get('user')
            password = kwargs.get('password')
            token = kwargs.get('token')

        # If you give a token, the user/password are ignored
        if token:
            if not isinstance(token, (str, bytes)):
                mqlog.trace_exit("qmgr:connect_with_options", ep=6)

                raise TypeError('Token must be an instance of str or bytes')

            csp._set_ptr_field('Token', token)

            # We need to fix these up even if they are not going to be used
            csp._set_ptr_field('CSPUserId', None)
            csp._set_ptr_field('CSPPassword', None)

            csp.AuthenticationType = CMQC.MQCSP_AUTH_ID_TOKEN
            csp.Version = max(csp.Version, CMQC.MQCSP_VERSION_3)

        elif user:
            # We check for None because password can be an empty string
            if password is None:
                mqlog.trace_exit("qmgr:connect_with_options", ep=7)
                raise ValueError('Password must not be None if user is provided')

            if not (isinstance(user, (str, bytes)) and isinstance(password, (str, bytes))):
                mqlog.trace_exit("qmgr:connect_with_options", ep=8)
                raise ValueError('Both user and password must be instances of str or bytes')
            csp._set_ptr_field('CSPUserId', user)
            csp._set_ptr_field('CSPPassword', password)
            csp.AuthenticationType = CMQC.MQCSP_AUTH_USER_ID_AND_PWD
        else:
            # Force these to be None
            csp._set_ptr_field('CSPUserId', None)
            csp._set_ptr_field('CSPPassword', None)

        if initial_key:
            csp._set_ptr_field('InitialKey', initial_key)
            csp.Version = max(csp.Version, CMQC.MQCSP_VERSION_2)

        if csp:
            csp_pack = csp.pack()

        if bno:
            bno_pack = bno.pack()

        cd_pack = None
        cd = kwargs['cd'] if 'cd' in kwargs else None
        if cd:

            # TLS encryption requires MQCD of version at least 7.
            # Thus, if someone uses TLS and the version is lower than that,
            # we can just increase it ourselves.
            if cd.SSLCipherSpec:
                cd.Version = max(cd.Version, CMQXC.MQCD_VERSION_7)

            ssl_peername = cd.SSLPeerNamePtr
            if ssl_peername:
                # Ideally we'd rename this field but have to keep its original
                # name for compatibility
                cd._set_ptr_field('SSLPeerNamePtr', ssl_peername)
                restore_cd = True

            cd_pack = cd.pack()

        # We are now able to call the real C function with all its parameters
        mqlog.debug(f"About to connect to {name}")
        rv = ibmmqc.MQCONNX(name, cno_pack, cd_pack, csp_pack, sco_pack, bno_pack)

        # The CSP/CNO structures might have been modified so restore the original parameters from
        # pointers to the actual string
        if restore_csp:
            csp.CSPUserId = user
            csp.CSPPassword = password
            if token:
                csp.Token = token
            if initial_key:
                csp.InitialKey = initial_key

        if restore_cno:
            if ccdt_url:
                cno.CCDTUrl = ccdt_url

        if restore_cd:
            if ssl_peername:
                cd.SSLPeerNamePtr = ssl_peername

        if restore_sco:
            if https_keystore:
                sco.HTTPSKeyStore = https_keystore
            if key_repo_password:
                sco.KeyRepoPassword = key_repo_password

        if rv[1] != CMQC.MQCC_FAILED:
            self.__handle = rv[0]
            self.__name = name

        if rv[1]:
            mqlog.trace_exit("qmgr:connect_with_options", ep=9, rc=rv[2])
            raise MQMIError(rv[1], rv[2])

        mqlog.trace_exit("qmgr:connect_with_options")

    # Backward compatibility
    connectWithOptions = connect_with_options

    # This is a basic connection for a client, where a default CD is used.
    # although it is possible to add further structures that get passed
    # down to the real CONNX operation. Many of the example programs come
    # via this method.
    def connect_tcp_client(self, name, cd, channel, conn_name,
                           user=None, password=None,
                           cno=None, csp=None, sco=None, bno=None
                           ):
        # type: (str, CD, str, str, Optional[str], Optional[str], Any, Any, Any, Any) -> None
        """ Connect immediately to the remote Queue Manager 'name', using
        a TCP Client connection, with channnel 'channel' and the
        TCP connection string 'conn_name'. Other connection
        options come from 'cd' and optionally the other connection-related
        classes.
        """

        mqlog.trace_entry("qmgr:connect_tcp_client")
        cd.ChannelName = ensure_strings_are_bytes(channel)
        cd.ConnectionName = ensure_strings_are_bytes(conn_name)
        cd.ChannelType = CMQXC.MQCHT_CLNTCONN
        cd.TransportType = CMQXC.MQXPT_TCP

        kwargs = {
            'user': user,
            'password': password,
            'opts': CMQC.MQCNO_CLIENT_BINDING,
            'cd': cd,
        }

        if cno:
            cno.Options |= CMQC.MQCNO_CLIENT_BINDING
            kwargs['cno'] = cno
        if csp:
            kwargs['csp'] = csp
        if sco:
            kwargs['sco'] = sco
        if bno:
            kwargs['bno'] = bno

        self.connect_with_options(name, **kwargs)
        mqlog.trace_exit("qmgr:connect_tcp_client")

    # Backward compatibility
    connectTCPClient = connect_tcp_client

    def disconnect(self) -> None:
        """ Disconnect from queue manager, if connected.
        """
        mqlog.trace_entry("qmgr:disconnect")

        if not self.__handle:
            mqlog.trace_exit("qmgr:disconnect", ep=1)
            raise PYIFError('not connected')
        saved_handle = self.__handle

        if OTelFunctions.disc:
            OTelFunctions.disc(self)

        rv = ibmmqc.MQDISC(self.__handle)
        if rv[0]:
            mqlog.trace_exit("qmgr:disconnect", ep=2, rc=rv[2])
            raise MQMIError(rv[0], rv[1])
        self.__handle = self.__qmobj = None
        mqcallback._delete_all_callbacks(saved_handle)
        mqlog.trace_exit("qmgr:disconnect")

    def get_handle(self) -> int:
        """ Get the queue manager handle. The handle is used for most other MQI calls.
        We don't trace this function in normal path as it would make things too verbose
        """
        if self.__handle:
            return self.__handle
        mqlog.trace("qmgr:get_handle raising PYIFError")
        raise PYIFError('not connected')

    def get_name(self) -> str:
        """ Get the queue manager name that was used during MQCONN(X).
        """
        if self.__name:
            v = self.__name
            return self.to_string(v)
        raise PYIFError('not connected')

    # Backward compatibility
    getHandle = get_handle

    def begin(self) -> None:
        """ Begin a new global transaction.
        """
        mqlog.trace_entry("qmgr:begin")

        rv = ibmmqc.MQBEGIN(self.__handle)
        if rv[0]:
            mqlog.trace_exit("qmgr:begin", ep=1, rc=rv[1])
            raise MQMIError(rv[0], rv[1])
        mqlog.trace_exit("qmgr:begin")

    def commit(self) -> None:
        """ Commits any outstanding gets/puts in the current unit of work.
        """
        mqlog.trace_entry("qmgr:commit")

        rv = ibmmqc.MQCMIT(self.__handle)
        if rv[0]:
            mqlog.trace_exit("qmgr:commit", ep=1, rc=rv[1])
            raise MQMIError(rv[0], rv[1])
        mqlog.trace_exit("qmgr:commit")

    def backout(self) -> None:
        """ Backout any outstanding gets/puts in the current unit of work.
        """
        mqlog.trace_entry("qmgr:backout")

        rv = ibmmqc.MQBACK(self.__handle)
        if rv[0]:
            mqlog.trace_exit("qmgr:backout", ep=1, rc=rv[1])
            raise MQMIError(rv[0], rv[1])
        mqlog.trace_exit("qmgr:backout")

    def put1(self, q_desc: Union[str, bytes, OD], msg: Optional[bytes], *opts: Union[MD, OD]) -> None:
        """ Put the single message in string buffer 'msg' on the queue
        using the MQI PUT1 call. This encapsulates calls to MQOPEN,
        MQPUT and MQCLOSE. put1 is the optimal way to put a single
        message on a queue.

        q_desc identifies the Queue either by name (if its a string),
        or by MQOD (if its a OD instance).

        mDesc is the MQMD Message Descriptor for the
        message. If it is not passed, or is None, then a default MD
        object is used.

        putOpts is the MQPMO Put Message Options structure
        for the put1 call. If it is not passed, or is None, then a
        default PMO object is used.

        If mDesc and/or putOpts arguments were supplied, they may be
        updated by the put1 operation.
        """

        mqlog.trace_entry("qmgr:put1")

        m_desc, put_opts = mqqargs.common_q_args(*opts)

        if not isinstance(msg, bytes):
            if isinstance(msg, str):  # Python 3 string is unicode
                msg = msg.encode(self.bytes_encoding)
                m_desc.CodedCharSetId = self.default_ccsid
                m_desc.Format = CMQC.MQFMT_STRING
            else:
                error_message = 'Message type is {0}. Convert to bytes.'
                mqlog.trace_exit("qmgr:put1", ep=1)
                raise TypeError(error_message.format(type(msg)))
        if put_opts is None:
            put_opts = PMO()

        if OTelFunctions.put_trace_before:
            OTelFunctions.put_trace_before(self, m_desc, put_opts, msg)

        # Now send the message
        rv = ibmmqc.MQPUT1(self.__handle, mqqargs._make_q_desc(q_desc).pack(), m_desc.pack(), put_opts.pack(), msg)
        if rv[-2]:
            mqlog.trace_exit("qmgr:put1", ep=1, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])
        _ = m_desc.unpack(rv[0])
        _ = put_opts.unpack(rv[1])

        if OTelFunctions.put_trace_after:
            OTelFunctions.put_trace_after(self, put_opts)
        mqlog.trace_exit("qmgr:put1")

    def inquire(self, selectors: Union[int, list[int]]) -> Union[Any, Dict[int, Any]]:
        """ Inquire on qmgr attributes. If the qmgr is not already
        open, it is opened for Inquire.

        If the selectors parameter is a single value, then that specific
        attribute's value is returned (string or int).

        If the selectors parameter is a list of values, then a dict is returned
        where all the values are stored using each element of the selectors as the keys.
        """
        mqlog.trace_entry("qmgr:inquire")

        if self.__qmobj is None:
            # Make an od for the queue manager, open the qmgr & cache result
            qmod = OD(ObjectType=CMQC.MQOT_Q_MGR)
            hdl = self.__handle if self.__handle else CMQC.MQHC_UNUSABLE_HCONN
            rv = ibmmqc.MQOPEN(hdl, qmod.pack(), CMQC.MQOO_INQUIRE)
            if rv[-2]:
                mqlog.trace_exit("qmgr:inquire", ep=1, rc=rv[-1])
                raise MQMIError(rv[-2], rv[-1])
            self.__qmobj = rv[0]
        # mqinq.inq will throw the exception if necessary
        rv = mqinq.common_inq(self.__handle, self.__qmobj, selectors)
        mqlog.trace_exit("qmgr:inquire")

        return rv

    # Create an alias that is closer to the real MQI function name
    inq = inquire

    def stat(self, status_type: int) -> STS:
        """Implementation of MQSTAT"""
        mqlog.trace_entry("qmgr:stat")

        stat = STS()
        rv = ibmmqc.MQSTAT(self.__handle, status_type, stat.pack())
        if rv[1]:
            mqlog.trace_exit("qmgr:stat", ep=1, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])
        mqlog.trace_exit("qmgr:stat")
        return stat.unpack(rv[0])

    def _is_connected(self) -> bool:
        """ Try accessing the qmgr to see whether the application
        is connected to it. Note that the method is merely a convenience wrapper around
        MQINQ(). In particular, there's still a possibility that
        the app will disconnect between checking QueueManager.is_connected
        and the next MQ call. And this operation might (silently) fail for other reasons such
        as an authorisation issue, giving the false impression that it is not connected.
        """

        # The original version of this function used the PCF PING_Q_MGR command, but
        # a) that requires more authorisations to be set
        # b) was a problem with circular imports when everything was split into multiple files
        # Using MQINQ is equally (un)reliable.
        try:
            self.inquire(CMQC.MQCA_Q_MGR_NAME)
            return True
        except MQMIError:
            return False

    is_connected = property(_is_connected)

    # Setting up a callback function
    # There is a similar method on the Queue object but they
    # both go to a common implementation
    def cb(self, **kwargs: Dict[str, Any]) -> None:
        """cb(operation=operation, md=MD,gmo=GMO,cbd=CBD)
        Register or Deregister a Callback function for asynchronous
        message consumption.

        The cbd.CallbackFunction must be defined as (Dict[str,Any]) with
        entries for queue_manager,queue(unless it's a qmgr-wide event),
        md,gmo,cbc,msg.
        """
        mqlog.trace_entry("qmgr:cb")
        mqcallback.real_cb(self, kwargs)
        mqlog.trace_exit("qmgr:cb")

    def ctl(self, operation: int, ctlo: CTLO) -> None:
        """Start or stop registered callbacks with the MQCTL operation.
        The connectionArea is stashed so it can be given to the callback
        function. That area is not removed until MQDISC but it might be
        overwritten by subsequent calls to this function.
        """
        mqlog.trace_entry("qmgr:ctl")

        if not isinstance(ctlo, CTLO):
            mqlog.trace_exit("qmgr:ctl", ep=1)
            raise TypeError("ctlo must be an instance of CTLO")

        mqcallback._save_connection_area(self.__handle, ctlo)
        original_cna = ctlo.ConnectionArea
        # Have to set this "pointer" field to NULL for pack() to work
        ctlo.ConnectionArea = 0
        rv = ibmmqc.MQCTL(self.__handle, operation, ctlo.pack())

        # But restore it immediately so the app doesn't notice the swap
        ctlo.ConnectionArea = original_cna
        if rv[1]:
            mqlog.trace_exit("qmgr:ctl", ep=2, rc=rv[-1])
            raise MQMIError(rv[-2], rv[-1])
        mqlog.trace_exit("qmgr:ctl")

# ################################################################################################################################

def connect(queue_manager, channel=None, conn_info=None, user=None, password=None, disconnect_on_exit=True,
            bytes_encoding=EncodingDefault.bytes_encoding, default_ccsid=EncodingDefault.ccsid,
            cd=None, cno=None, csp=None, sco=None, bno=None
            ):
    """ A convenience wrapper for connecting to MQ queue managers without needing to explicitly create a qmgr object first.
    If given both 'channel' and 'conn_info' will connect in client mode. If neither are given
    then a default connection mode is attempted. That might be either local bindings or a client,
    depending on other environmental factors.

    A QueueManager() is returned after successfully establishing a connection.
    """

    mqlog.trace_entry(":connect")
    qmgr = QueueManager(None, disconnect_on_exit, bytes_encoding=bytes_encoding, default_ccsid=default_ccsid)

    if channel and conn_info:
        qmgr.connect_tcp_client(queue_manager or '', cd or CD(), channel, conn_info, user, password, cno=cno, csp=csp, sco=sco, bno=bno)

    elif queue_manager:
        qmgr.connect_with_options(queue_manager, user=user, password=password, cno=cno, csp=csp, sco=sco, bno=bno, cd=cd)

    else:
        mqlog.trace_exit(":connect", ep=1)
        raise TypeError('Invalid arguments: %s' % repr([queue_manager, channel, conn_info, user, password]))

    mqlog.trace_exit(":connect")
    return qmgr



"""Message Handles and Message Properties"""

# Copyright (c) 2025,2026 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqerrors import *
from ibmmq import CMQC, ibmmqc
from mqsub import *
from mqprops import *

from mqqmgr import *
import mqlog

class MessageHandle:
    """ A higher-level wrapper around the MQI's native Message Handle and
    Property processing.
    """

    class _Properties:
        """ Encapsulates access to message properties.
        """

        def __init__(self, conn_handle, msg_handle):
            self.conn_handle = conn_handle
            self.msg_handle = msg_handle

            # When accessing message properties, this will be the maximum number
            # of characters a value will be able to hold. If it's not enough
            # an exception will be raised and its 'actual_value_length' will be
            # filled in with the information of how many characters there are actually
            # so that an application may re-issue the call.
            self.default_value_length = 64

        def __getitem__(self, name):
            """ Allows for a dict-like access to properties,
            handle.properties[name]
            """
            value = self.get(name)
            if not value:
                raise KeyError('No such property [%s]' % name)

            return value

        def __setitem__(self, name, value):
            """ Implements 'handle.properties[name] = value'.
            """
            return self.set(name, value)

        def get(self, name, default=None, max_value_length=None,
                impo_options=CMQC.MQIMPO_INQ_FIRST, impo=None, pd=None,
                property_type=CMQC.MQTYPE_AS_SET):
            """ Returns the value of message property 'name'. If a wildcard
            is used in the property name, then the real name is also returned.

            'max_value_length' is the maximum number of characters the underlying
            C function is allowed to allocate for fetching the value
            (defaults to default_value_length).

            Either 'impo_options' or 'impo' can be given. If the full IMPO
            object is used, then its options take preference.

            The 'pd' can be either the options value or the PD class. (The original parameter name
            did not include "options" so it's easier to repurpose it to take either type of
            value than it is wth the impo for compatibility.)

            'property_type' points to the expected data type of the property.

            The name can be '%' for a wildcard, and can be preceded by a
            folder. For example, 'usr.%'. But you cannot use, say, 'ABC%'. The
            wildcard means all properties within a single folder.
            """
            mqlog.trace_entry("msghdl:get")
            if impo:
                if not isinstance(impo, IMPO):
                    mqlog.trace_exit("msghdl:get", ep=1)
                    raise TypeError("impo must be an instance of IMPO")
            else:
                impo = IMPO()
                impo.Options = impo_options | CMQC.MQIMPO_CONVERT_VALUE

            if pd:
                pd_val = pd
                if isinstance(pd_val, int):
                    pd = PD()
                    pd.Options = pd
                if not isinstance(pd, PD):
                    mqlog.trace_exit("msghdl:get", ep=2)
                    raise TypeError("pd must be an instance of PD")
            else:
                pd = PD()

            name = ensure_strings_are_bytes(name)

            if not max_value_length:
                max_value_length = self.default_value_length

            value, data_length, returned_name, comp_code, comp_reason = ibmmqc.MQINQMP(
                self.conn_handle, self.msg_handle, impo.pack(), name, pd.pack(), property_type, max_value_length)

            if comp_code != CMQC.MQCC_OK:
                mqlog.trace_exit("msghdl:get", ep=2, rc=comp_reason)
                raise MQMIError(comp_code, comp_reason, value=default, data_length=data_length)

            if returned_name:
                mqlog.trace_exit("msghdl:get", ep=3)
                return value, returned_name
            mqlog.trace_exit("msghdl:get")
            return value

        def set(self, name, value, property_type=CMQC.MQTYPE_STRING,
                value_length=CMQC.MQVL_NULL_TERMINATED, pd=None, smpo=None):
            """ Allows for setting arbitrary properties of a message. 'name'
            and 'value' are mandatory. All other parameters are OK to use as-is
            if 'value' is a string. If it isn't a string, the 'property_type'
            and 'value_length' should be set accordingly. For further
            customization, you can also use 'pd' and 'smpo' parameters for
            passing in MQPD and MQSMPO structures.
            """
            mqlog.trace_entry("msghdl:set")

            name = ensure_strings_are_bytes(name)

            # If the VALUE is of MQTYPE_STRING, then the input is expected to be a real string
            # (Unicode allowed) and it's converted in the C layer. Unlike the name, we do not convert
            # it here.

            pd = pd if pd else PD()
            smpo = smpo if smpo else SMPO()

            comp_code, comp_reason = ibmmqc.MQSETMP(
                self.conn_handle, self.msg_handle, smpo.pack(), name, pd.pack(), property_type, value, value_length)

            if comp_code != CMQC.MQCC_OK:
                mqlog.trace_exit("msghdl:set", ep=1, rc=comp_reason)
                raise MQMIError(comp_code, comp_reason)
            mqlog.trace_exit("msghdl:set")

        def dlt(self, name, dmpo=None):
            """ Deletes a message property. Only the name is required. For further
            customization, you can also use the 'dmpo' parameters for
            passing in the MQDMPO structure.
            """
            mqlog.trace_entry("msghdl:dltproperty")

            name = ensure_strings_are_bytes(name)

            dmpo = dmpo if dmpo else DMPO()

            comp_code, comp_reason = ibmmqc.MQDLTMP(
                self.conn_handle, self.msg_handle, dmpo.pack(), name)

            if comp_code != CMQC.MQCC_OK:
                mqlog.trace_exit("msghdl:dlt", ep=1, rc=comp_reason)
                raise MQMIError(comp_code, comp_reason)
            mqlog.trace_exit("msghdl:dltproperty")

    def __init__(self, qmgr=None, cmho=None, dup_handle=None):
        """There may be times when we need to create a message handle object
        when all we have is the integer handle value (the PMO/GMO fields are
        set to the integers, not objects). So we can't use those directly
        to set/inquire additional properties. The dup_handle field lets us
        set the properties using this new object that is still the same
        thing in the underlying C code.
        """
        mqlog.trace_entry("msghdl:__init__")

        self.conn_handle = qmgr.get_handle() if qmgr else CMQC.MQHO_NONE
        cmho = cmho if cmho else CMHO()

        if dup_handle is None:
            self.msg_handle, comp_code, comp_reason = ibmmqc.MQCRTMH(self.conn_handle, cmho.pack())

            if comp_code != CMQC.MQCC_OK:
                mqlog.trace_exit("msghdl:__init__", ep=1, rc=comp_reason)
                raise MQMIError(comp_code, comp_reason)
        else:
            self.msg_handle = dup_handle

        self.properties = self._Properties(self.conn_handle, self.msg_handle)
        mqlog.trace_exit("msghdl:__init__")

    def get_handle(self):
        """Get the actual integer value"""
        return self.msg_handle

    def get_queue_manager(self):
        """Get the associated hConn"""
        return self.conn_handle

    # Note that this deletes a MsgHandle at the MQI level and is not __del__ (the object destructor)
    def dlt(self, dmho=None):
        """Delete a message handle"""
        mqlog.trace_entry("msghdl:dlthdl")
        dmho = dmho if dmho else DMHO()

        comp_code, comp_reason = ibmmqc.MQDLTMH(self.conn_handle, self.msg_handle, dmho.pack())

        if comp_code != CMQC.MQCC_OK:
            mqlog.trace_exit("msghdl:dlthdl", ep=1, rc=comp_reason)
            raise MQMIError(comp_code, comp_reason)

        self.properties = self._Properties(CMQC.MQHC_UNUSABLE_HCONN, CMQC.MQHM_NONE)
        mqlog.trace_exit("msghdl:dlthdl")



""" All sorts of tests related to making the API PEP-8 compliant.
"""
import unittest

import ibmmq as mq

class TestApiTransitionPEP8(unittest.TestCase):
    """All sorts of tests related to making the API PEP-8 compliant."""

    def test_backward_compatibility(self):
        """Test backward-compatible.

        Makes sure all the relevant classes and methods have
        backward-compatible replacements.
        """
        self.assertEqual(mq.gmo, mq.GMO)
        self.assertEqual(mq.pmo, mq.PMO)
        self.assertEqual(mq.od, mq.OD)
        self.assertEqual(mq.md, mq.MD)
        self.assertEqual(mq.cd, mq.CD)
        self.assertEqual(mq.sco, mq.SCO)
        self.assertEqual(mq.QueueManager.connectWithOptions, mq.QueueManager.connect_with_options)
        self.assertEqual(mq.QueueManager.connectTCPClient, mq.QueueManager.connect_tcp_client)
        self.assertEqual(mq.QueueManager.getHandle, mq.QueueManager.get_handle)
        self.assertEqual(mq.PCFExecute.stringifyKeys, mq.PCFExecute.stringify_keys)



"""MQPCF structures: CFH, CFIN, CFIN64 etc"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, CMQCFC

try:
    from typing import Any, Dict, Union
except ImportError:
    pass

class CFH(MQOpts):
    """ Construct an MQCFH Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw: Dict[str, Any]):
        opts = [['Type', CMQCFC.MQCFT_COMMAND, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFH_STRUC_LENGTH, MQLONG_TYPE],
                ['Version', CMQCFC.MQCFH_VERSION_3, MQLONG_TYPE],
                ['Command', CMQCFC.MQCMD_NONE, MQLONG_TYPE],
                ['MsgSeqNumber', 1, MQLONG_TYPE],
                ['Control', CMQCFC.MQCFC_LAST, MQLONG_TYPE],
                ['CompCode', CMQC.MQCC_OK, MQLONG_TYPE],
                ['Reason', CMQC.MQRC_NONE, MQLONG_TYPE],
                ['ParameterCount', 0, MQLONG_TYPE],
                ]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFBF(MQOpts):
    """ Construct an MQCFBF (PCF Byte String Filter) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """

    def __init__(self, **kw: Dict[str, Any]):
        filter_value = kw.pop('FilterValue', '')
        filter_value_length = kw.pop('FilterValueLength', len(filter_value))  # type: int
        padded_filter_value_length = padded_count(filter_value_length)

        opts = [['Type', CMQCFC.MQCFT_BYTE_STRING_FILTER, MQLONG_TYPE],
                ['StrucLength',
                 CMQCFC.MQCFBF_STRUC_LENGTH_FIXED + padded_filter_value_length, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Operator', 0, MQLONG_TYPE],
                ['FilterValueLength', filter_value_length, MQLONG_TYPE],
                ['FilterValue', filter_value, '{}s'.format(padded_filter_value_length)]
                ]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFBS(MQOpts):
    """ Construct an MQCFBS (PCF Byte String) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """

    def __init__(self, **kw: Dict[str, Any]):
        string = kw.pop('String', '')
        string_length = kw.pop('StringLength', len(string))  # type: int
        padded_string_length = padded_count(string_length)

        opts = [['Type', CMQCFC.MQCFT_BYTE_STRING, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFBS_STRUC_LENGTH_FIXED + padded_string_length, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['StringLength', string_length, MQLONG_TYPE],
                ['String', string, '{}s'.format(padded_string_length)]
                ]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFGR(MQOpts):
    """ Construct an MQCFGR (PCF Group) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """

    def __init__(self, **kw: Dict[str, Any]):
        count = kw.pop('ParameterCount', 0)

        opts = [['Type', CMQCFC.MQCFT_GROUP, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFGR_STRUC_LENGTH, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['ParameterCount', count, MQLONG_TYPE],
                ]
        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFIF(MQOpts):
    """ Construct an MQCFIF (PCF Integer Filter) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """

    def __init__(self, **kw: Dict[str, Any]):
        opts = [['Type', CMQCFC.MQCFT_INTEGER_FILTER, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFIF_STRUC_LENGTH, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Operator', 0, MQLONG_TYPE],
                ['FilterValue', 0, MQLONG_TYPE]
                ]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFIL(MQOpts):
    """ Construct an MQCFIL (PCF 32-bit integer List) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw: Dict[str, Any]):
        values = kw.pop('Values', [])  # type: list[int]
        count = kw.pop('Count', len(values))  # type: int

        opts = [['Type', CMQCFC.MQCFT_INTEGER_LIST, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFIL_STRUC_LENGTH_FIXED + 4 * count, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Count', count, MQLONG_TYPE],
                ['Values', values, MQLONG_TYPE, count],
                ]
        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFIL64(MQOpts):
    """ Construct an MQCFIL64 (PCF 64-bit integer List) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw: Dict[str, Any]):
        values = kw.pop('Values', [])  # type: list[int]
        count = kw.pop('Count', len(values))  # type: int

        opts = [['Type', CMQCFC.MQCFT_INTEGER64_LIST, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFIL64_STRUC_LENGTH_FIXED + 8 * count, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Count', count, MQLONG_TYPE],
                ['Values', values, INTEGER64_TYPE, count],
                ]
        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFIN(MQOpts):
    """ Construct an MQCFIN (PCF 32-bit integer) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw: Dict[str, Any]):
        opts = [['Type', CMQCFC.MQCFT_INTEGER, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFIN_STRUC_LENGTH, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Value', 0, MQLONG_TYPE],
                ]
        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFIN64(MQOpts):
    """ Construct an MQCFIN64 (PCF 64-bit integer) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw: Dict[str, Any]):
        opts = [['Type', CMQCFC.MQCFT_INTEGER64, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFIN64_STRUC_LENGTH, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Value', 0, INTEGER64_TYPE],
                ]
        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFSF(MQOpts):
    """ Construct an MQCFSF (PCF String Filter) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """

    def __init__(self, **kw: Dict[str, Any]):
        filter_value = kw.pop('FilterValue', '')
        filter_value_length = kw.pop('FilterValueLength', len(filter_value))  # type: int
        padded_filter_value_length = padded_count(filter_value_length)

        opts = [['Type', CMQCFC.MQCFT_STRING_FILTER, MQLONG_TYPE],
                ['StrucLength',
                 CMQCFC.MQCFSF_STRUC_LENGTH_FIXED + padded_filter_value_length, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['Operator', 0, MQLONG_TYPE],
                ['CodedCharSetId', CMQC.MQCCSI_DEFAULT, MQLONG_TYPE],
                ['FilterValueLength', filter_value_length, MQLONG_TYPE],
                ['FilterValue', filter_value, '{}s'.format(padded_filter_value_length)]
                ]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFSL(MQOpts):
    """ Construct an MQCFSL (PCF String List) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """

    def __init__(self, **kw: Dict[str, Any]):
        strings = kw.pop('Strings', [])  # type: list[Union[str,bytes]]
        string_length = kw.pop('StringLength', len(max(strings, key=len)) if strings else 0)  # type: int

        strings_count = len(strings)
        count = kw.pop('Count', strings_count)  # type: int

        max_string_length = padded_count(string_length) if count else 0
        padded_strings_length = (max_string_length) * strings_count

        opts = [['Type', CMQCFC.MQCFT_STRING_LIST, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFSL_STRUC_LENGTH_FIXED + padded_strings_length, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['CodedCharSetId', CMQC.MQCCSI_DEFAULT, MQLONG_TYPE],
                ['Count', count, MQLONG_TYPE],
                ['StringLength', max_string_length, MQLONG_TYPE],
                ['Strings', strings if strings else [b''], '{}s'.format(max_string_length), (count if count else 1)]
                ]

        super().__init__(tuple(opts), **kw)

# ################################################################################################################################

class CFST(MQOpts):
    """ Construct an MQCFST (PCF String) structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    Note that the "String" may include the padding bytes, so you have to use the
    StringLength field to extract the real value.
    """

    def __init__(self, **kw: Dict[str, Any]):
        string = kw.pop('String', '')
        string_length = kw.pop('StringLength', len(string))  # type: int
        padded_string_length = padded_count(string_length)

        opts = [['Type', CMQCFC.MQCFT_STRING, MQLONG_TYPE],
                ['StrucLength', CMQCFC.MQCFST_STRUC_LENGTH_FIXED + padded_string_length, MQLONG_TYPE],
                ['Parameter', 0, MQLONG_TYPE],
                ['CodedCharSetId', CMQC.MQCCSI_DEFAULT, MQLONG_TYPE],
                ['StringLength', string_length, MQLONG_TYPE],
                ['String', string, '{}s'.format(padded_string_length)]
                ]

        super().__init__(tuple(opts), **kw)



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example demonstrates the "is_connected" property of a qmgr.
"""

import logging

import ibmmq as mq

logging.basicConfig(level=logging.INFO)

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)
logging.info('qmgr.is_connected=`%s`', qmgr.is_connected)

qmgr.disconnect()



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example demonstrates setting and viewing the priority of a message. A
random number in the valid range is set in the message.
"""

import logging
import random
import ibmmq as mq

logging.basicConfig(level=logging.INFO)

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
message = 'Hello from Python!'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

priority = random.randint(0, 9)

put_md = mq.MD()
put_md.Priority = priority

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

put_queue = mq.Queue(qmgr, queue_name)
put_queue.put(message, put_md)

get_md = mq.MD()
get_queue = mq.Queue(qmgr, queue_name)
message_body = get_queue.get(None, get_md)

logging.info('Received a message, priority `%s`.', get_md.Priority)

put_queue.close()
get_queue.close()
qmgr.disconnect()

print("Done.")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows a simple execution of a PCF command to display local queues matching a pattern
and then show their current depth.

The PCF parameters are built as a list of separate elements. Compare with the the dict-based approach
in the dis_channels example.

The "admin" credentials are used to avoid needing to grant application users
additional privileges. We use the CSP class to provide those credentials.
"""

import ibmmq as mq

no_queues = False
queue_manager = 'QM1'
channel = 'DEV.ADMIN.SVRCONN'
host = '127.0.0.1'
port = '1414'

conn_info = '%s(%s)' % (host, port)

csp = mq.CSP()
csp.CSPUserId = 'admin'
csp.CSPPassword = 'password'

prefix = 'SYSTEM.*'
queue_type = mq.CMQC.MQQT_LOCAL

# The parameters needed for the INQUIRE Q command.
name_attrs = []
name_attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_Q_NAME,
                          String=prefix))
name_attrs.append(mq.CFIN(Parameter=mq.CMQC.MQIA_Q_TYPE,
                          Value=queue_type))

# The parameters needed for the INQUIRE Q STATUS command
status_attrs = []
status_attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_Q_NAME, String=prefix))
status_attrs.append(mq.CFIL(Parameter=mq.CMQCFC.MQIACF_Q_STATUS_ATTRS, Values=[mq.CMQCFC.MQIACF_ALL]))

qmgr = mq.connect(queue_manager, channel, conn_info, csp=csp)
pcf = mq.PCFExecute(qmgr, response_wait_interval=15000)

print()
print("Listing queues...")
try:
    response = pcf.MQCMD_INQUIRE_Q(name_attrs)
except mq.MQMIError as e:
    if e.comp == mq.CMQC.MQCC_FAILED and e.reason == mq.CMQC.MQRC_UNKNOWN_OBJECT_NAME:
        no_queues = True
        print('No queues matched given pattern.')
    else:
        raise
else:
    for queue_info in response:
        try:
            queue_name = queue_info[mq.CMQC.MQCA_Q_NAME]
            print(f'Found queue \"{mq.to_string(queue_name)}\"')
        except UnicodeError:
            print(f'Failure to decode msg: {queue_name}')
        except KeyError:
            print('Cannot find queue name in response: ', queue_info)

if not no_queues:
    print()
    print("Listing queue status...")
    try:
        response = pcf.MQCMD_INQUIRE_Q_STATUS(status_attrs)
    except mq.MQMIError as e:
        if e.comp == mq.CMQC.MQCC_FAILED and e.reason == mq.CMQC.MQRC_UNKNOWN_OBJECT_NAME:
            # Should not get here as we already know there are some queues matching the pattern
            print('No queues matched given pattern.')
        else:
            raise
    else:
        for queue_info in response:
            try:
                queue_name = queue_info[mq.CMQC.MQCA_Q_NAME]
                depth = queue_info[mq.CMQC.MQIA_CURRENT_Q_DEPTH]
                lputtime = mq.to_string(queue_info[mq.CMQCFC.MQCACF_LAST_PUT_TIME])
                lputdate = mq.to_string(queue_info[mq.CMQCFC.MQCACF_LAST_PUT_DATE])
                if lputdate != "":
                    lput = lputdate + ":" + lputtime
                else:
                    lput = "N/A"
                print(f'Found queue {queue_name} depth: {depth} lastPut: {lput}')
            except KeyError as e:
                print('Failure to decode msg because ', e)

qmgr.disconnect()

print("Done.")



"""Test setting message properties and getting its values
"""
import unittest

import ibmmq as mq

import config
import utils

# pylint: disable=too-many-instance-attributes,missing-function-docstring

class TestMP(unittest.TestCase):
    """Testing Message Properties"""
    def setUp(self):

        self.msg_prop_name = b"test_name"

        self.msg_prop_value_str = "test_valuetest_valuetest_valuetest_valuetest_value"
        self.msg_prop_value_bytes = b"test_valuetest_valuetest_valuetest_valuetest_value"
        self.msg_prop_value_bool = True
        self.msg_prop_value_int8 = -127
        self.msg_prop_value_int16 = -32768
        self.msg_prop_value_int32 = -2147483647
        self.msg_prop_value_int64 = -9223372036854775808
        self.msg_prop_value_float32 = 1.1754943508222875e-38
        self.msg_prop_value_float64 = 2.2250738585072014e-308

        # max length of queue names is 48 characters
        self.queue_name = "{prefix}MSG.PROP.QUEUE".format(prefix=config.MQ.QUEUE.PREFIX)
        self.queue_manager = config.MQ.QM.NAME
        self.channel = config.MQ.QM.CHANNEL
        self.host = config.MQ.QM.HOST
        self.port = config.MQ.QM.PORT
        self.user = config.MQ.QM.USER
        self.password = config.MQ.QM.PASSWORD

        self.conn_info = "{0}({1})".format(self.host, self.port)

        self.qmgr = mq.QueueManager(None)
        self.qmgr.connect_tcp_client(self.queue_manager, mq.CD(), self.channel, self.conn_info, self.user, self.password)

        self.create_queue(self.queue_name)

    def tearDown(self):
        """Delete the created objects.
        """
        if self.queue_name:
            self.delete_queue(self.queue_name)
        self.qmgr.disconnect()

    def create_queue(self, queue_name):
        """Create the queue we'll use for testing against"""
        queue_type = mq.CMQC.MQQT_LOCAL
        max_depth = 5000

        args = {mq.CMQC.MQCA_Q_NAME: utils.py3str2bytes(queue_name),
                mq.CMQC.MQIA_Q_TYPE: queue_type,
                mq.CMQC.MQIA_MAX_Q_DEPTH: max_depth,
                mq.CMQCFC.MQIACF_REPLACE: mq.CMQCFC.MQRP_YES}
        pcf = mq.PCFExecute(self.qmgr, response_wait_interval=120000)
        pcf.MQCMD_CREATE_Q(args)
        pcf.disconnect()

    def delete_queue(self, queue_name):
        """Delete the queue"""
        pcf = mq.PCFExecute(self.qmgr, response_wait_interval=120000)
        args = {mq.CMQC.MQCA_Q_NAME: utils.py3str2bytes(queue_name),
                mq.CMQCFC.MQIACF_PURGE: mq.CMQCFC.MQPO_YES}
        pcf.MQCMD_DELETE_Q(args)

    def get_value_length(self, property_type, property_value=''):
        """How many bytes does each property type use"""
        value_length = 0
        if property_type == mq.CMQC.MQTYPE_BOOLEAN:
            value_length = 4
        elif property_type == mq.CMQC.MQTYPE_BYTE_STRING:
            value_length = len(property_value)
        elif property_type == mq.CMQC.MQTYPE_INT8:
            value_length = 1
        elif property_type == mq.CMQC.MQTYPE_INT16:
            value_length = 2
        elif property_type == mq.CMQC.MQTYPE_INT32:
            value_length = 4
        elif property_type == mq.CMQC.MQTYPE_INT64:
            value_length = 8
        elif property_type == mq.CMQC.MQTYPE_FLOAT32:
            value_length = 4
        elif property_type == mq.CMQC.MQTYPE_FLOAT64:
            value_length = 8
        elif property_type == mq.CMQC.MQTYPE_STRING:
            value_length = len(property_value)
        elif property_type == mq.CMQC.MQTYPE_NULL:
            value_length = 0

        return value_length

    def work_with_property(self, property_value, property_type):
        message_hdl_get = None
        queue_get = None
        queue_put = None
        try:

            value_length = self.get_value_length(property_type, property_value)

            cmho_put = mq.CMHO()
            message_hdl_put = mq.MessageHandle(self.qmgr, cmho_put)
            message_hdl_put.properties.set(self.msg_prop_name, property_value,
                                            value_length=value_length,
                                            property_type=property_type)

            pmo = mq.PMO(Version=mq.CMQC.MQPMO_CURRENT_VERSION)
            pmo.OriginalMsgHandle = message_hdl_put.msg_handle

            md_put = mq.MD(Version=mq.CMQC.MQMD_CURRENT_VERSION)

            queue_put = mq.Queue(self.qmgr, self.queue_name, mq.CMQC.MQOO_OUTPUT)
            queue_put.put(b'', md_put, pmo)

            queue_put.close()

            gmo = mq.GMO(Version=mq.CMQC.MQGMO_CURRENT_VERSION)
            gmo.Options = mq.CMQC.MQGMO_NO_WAIT | mq.CMQC.MQGMO_PROPERTIES_IN_HANDLE
            gmo.MatchOptions = mq.CMQC.MQMO_MATCH_MSG_ID

            cmho_get = mq.CMHO(Version=mq.CMQC.MQCMHO_CURRENT_VERSION)
            message_hdl_get = mq.MessageHandle(self.qmgr, cmho_get)
            gmo.MsgHandle = message_hdl_get.msg_handle
            md_get = mq.MD()
            md_get.MsgId = md_put.MsgId

            queue_get = mq.Queue(self.qmgr, self.queue_name, mq.CMQC.MQOO_INPUT_AS_Q_DEF)
            queue_get.get(None, md_get, gmo)
        finally:
            if queue_put:
                if queue_put.get_handle():
                    queue_put.close()

            if queue_get:
                if queue_get.get_handle():
                    queue_get.close()

        return message_hdl_get

    def get_property_value(self, message_hdl_get,
                           property_type=mq.CMQC.MQTYPE_AS_SET,
                           value_length=None,
                           property_modify=False):
        if not value_length:
            value_length = self.get_value_length(property_type)
        if property_modify:
            property_name = self.msg_prop_name * 2
        else:
            property_name = self.msg_prop_name

        return message_hdl_get.properties.get(property_name,
                                                property_type=property_type,
                                                max_value_length=value_length)

############################################################################
#
# Real Tests start here
#
############################################################################

    def test_message_properties_short(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_bytes, mq.CMQC.MQTYPE_BYTE_STRING)

        try:
            # Use "//" to always get an int value
            message_hdl_get.properties.get(self.msg_prop_name, max_value_length=len(self.msg_prop_value_bytes) // 2)
        except mq.MQMIError as e:
            self.assertEqual(e.reason, mq.CMQC.MQRC_PROPERTY_VALUE_TOO_BIG, e)

    def test_message_properties_byte(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_bytes, mq.CMQC.MQTYPE_BYTE_STRING)

        value = message_hdl_get.properties.get(self.msg_prop_name, max_value_length=len(self.msg_prop_value_bytes))

        self.assertEqual(self.msg_prop_value_bytes, value)

    def test_message_properties_str(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_str, mq.CMQC.MQTYPE_STRING)

        value = message_hdl_get.properties.get(self.msg_prop_name, max_value_length=len(self.msg_prop_value_str))

        self.assertEqual(self.msg_prop_value_str, value)

    def test_message_properties_bool(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_bool, mq.CMQC.MQTYPE_BOOLEAN)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_bool, value)

    def test_message_properties_int8(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_int8, mq.CMQC.MQTYPE_INT8)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_int8, value)

    def test_message_properties_int16(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_int16, mq.CMQC.MQTYPE_INT16)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_int16, value)

    def test_message_properties_int32(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_int32, mq.CMQC.MQTYPE_INT32)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_int32, value)

    def test_message_properties_int64(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_int64, mq.CMQC.MQTYPE_INT64)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_int64, value)

    def test_message_properties_float32(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_float32, mq.CMQC.MQTYPE_FLOAT32)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_float32, value)

    def test_message_properties_float64(self):
        message_hdl_get = self.work_with_property(self.msg_prop_value_float64, mq.CMQC.MQTYPE_FLOAT64)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(self.msg_prop_value_float64, value)

    def test_message_properties_null(self):
        message_hdl_get = self.work_with_property(None, mq.CMQC.MQTYPE_NULL)
        value = self.get_property_value(message_hdl_get)

        self.assertEqual(None, value)

    def test_message_properties_nonexist(self):
        message_hdl_get = self.work_with_property(None, mq.CMQC.MQTYPE_NULL)
        try:
            _ = self.get_property_value(message_hdl_get, property_modify=True)

        except mq.MQMIError as e:
            self.assertEqual(e.reason, mq.CMQC.MQRC_PROPERTY_NOT_AVAILABLE, e)


if __name__ == "__main__":
    unittest.main()



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows how to put a message with an alternate userid. It will
fail (2035) if the real user does not have the authority to use the option, and
if the alternate user does not have the authority to put to the queue.
"""
import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
message = 'Hello from Python!'
alternate_user_id = 'altusr'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

od = mq.OD()
od.ObjectName = queue_name
od.AlternateUserId = alternate_user_id

queue = mq.Queue(qmgr)
queue.open(od, mq.CMQC.MQOO_OUTPUT | mq.CMQC.MQOO_ALTERNATE_USER_AUTHORITY)
queue.put(message)

queue.close()
qmgr.disconnect()

print("Done.")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example browses a queue, formatting any Dead Letter Headers that it finds
"""

import ibmmq as mq
from ibmmq import CMQC

# Use a local connection instead of client
qmgr = mq.connect('QM1')

ok = True

od = mq.OD()
od.ObjectName = "SYSTEM.DEAD.LETTER.QUEUE"
q = mq.Queue(qmgr, od, CMQC.MQOO_BROWSE)

gmo = mq.GMO()
gmo.Options = CMQC.MQGMO_BROWSE_FIRST
while ok:
    try:
        md = mq.MD()
        msg = q.get(None, md, gmo)
        gmo.Options = CMQC.MQGMO_BROWSE_NEXT
        fmt = md['Format']

        print("----------------------")
        if fmt == CMQC.MQFMT_DEAD_LETTER_HEADER:
            offset = CMQC.MQDLH_LENGTH_1

            # Print a neatly formatted version of the DLH - strings are real strings
            dlh = mq.DLH().get_header(msg)

            # The to_string method changes the DLH contents, so we stash the
            # Format in its bytes version
            fmt = dlh['Format']
            print("DLH: ", dlh.to_string())

            # If there is an RFH2, then also decode and print that. The encoding
            # for the RFH2 (big/little-endian indicator) comes from the DLH.
            if fmt == CMQC.MQFMT_RF_HEADER_2:
                rfh2 = mq.RFH2()
                rfh2.unpack(msg[offset:], dlh['Encoding'])
                print()
                print("RFH: ", rfh2.to_string())
                offset += rfh2['StrucLength']

            # And now print the message body
            print()
            print("MSG: ", msg[offset:])
        else:
            print(msg)
        print()

    except mq.MQMIError as e:
        if e.reason == CMQC.MQRC_NO_MSG_AVAILABLE:
            print("No more messages.")
            ok = False
        else:
            raise
q.close()
qmgr.disconnect()



""""MQCD: Channel Descriptor"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, CMQXC, ibmmqc

# MQCONNX code courtesy of John OSullivan (mailto:jos@onebox.com)
# SSL additions courtesy of Brian Vicente (mailto:sailbv@netscape.net)

class CD(MQOpts):
    """ Construct an MQCD Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    The VERSION_11 structure was current from at least MQ 8.0 and does not need to flow
    to other systems, so we take that as the default version.
    """
    def __init__(self, **kw):
        """__init__(**kw)"""
        opts = []
        opts += [
            ['ChannelName', b'', '20s'],
            ['Version', CMQXC.MQCD_VERSION_11, MQLONG_TYPE],
            ['ChannelType', CMQXC.MQCHT_CLNTCONN, MQLONG_TYPE],
            ['TransportType', CMQXC.MQXPT_TCP, MQLONG_TYPE],
            ['Desc', b'', '64s'],
            ['QMgrName', b'', '48s'],
            ['XmitQName', b'', '48s'],
            ['ShortConnectionName', b'', '20s'],
            ['MCAName', b'', '20s'],
            ['ModeName', b'', '8s'],
            ['TpName', b'', '64s'],
            ['BatchSize', (50), MQLONG_TYPE],
            ['DiscInterval', (6000), MQLONG_TYPE],
            ['ShortRetryCount', (10), MQLONG_TYPE],
            ['ShortRetryInterval', (60), MQLONG_TYPE],
            ['LongRetryCount', (999999999), MQLONG_TYPE],
            ['LongRetryInterval', (1200), MQLONG_TYPE],
            ['SecurityExit', b'', str(CMQC.MQ_EXIT_NAME_LENGTH) + 's'],
            ['MsgExit', b'', str(CMQC.MQ_EXIT_NAME_LENGTH) + 's'],
            ['SendExit', b'', str(CMQC.MQ_EXIT_NAME_LENGTH) + 's'],
            ['ReceiveExit', b'', str(CMQC.MQ_EXIT_NAME_LENGTH) + 's'],
            ['SeqNumberWrap', (999999999), MQLONG_TYPE],
            ['MaxMsgLength', (4194304), MQLONG_TYPE],
            ['PutAuthority', CMQXC.MQPA_DEFAULT, MQLONG_TYPE],
            ['DataConversion', CMQXC.MQCDC_NO_SENDER_CONVERSION, MQLONG_TYPE],
            ['SecurityUserData', b'', '32s'],
            ['MsgUserData', b'', '32s'],
            ['SendUserData', b'', '32s'],
            ['ReceiveUserData', b'', '32s'],
            # Version 1
            ['UserIdentifier', b'', '12s'],
            ['Password', b'', '12s'],
            ['MCAUserIdentifier', b'', '12s'],
            ['MCAType', CMQXC.MQMCAT_PROCESS, MQLONG_TYPE],
            ['ConnectionName', b'', '264s'],
            ['RemoteUserIdentifier', b'', '12s'],
            ['RemotePassword', b'', '12s'],
            # Version 2
            ['MsgRetryExit', b'', str(CMQC.MQ_EXIT_NAME_LENGTH) + 's'],
            ['MsgRetryUserData', b'', '32s'],
            ['MsgRetryCount', (10), MQLONG_TYPE],
            ['MsgRetryInterval', (1000), MQLONG_TYPE],
            # Version 3
            ['HeartbeatInterval', (300), MQLONG_TYPE],
            ['BatchInterval', (0), MQLONG_TYPE],
            ['NonPersistentMsgSpeed', CMQXC.MQNPMS_FAST, MQLONG_TYPE],
            ['StrucLength', CMQXC.MQCD_CURRENT_LENGTH, MQLONG_TYPE],
            ['ExitNameLength', CMQC.MQ_EXIT_NAME_LENGTH, MQLONG_TYPE],
            ['ExitDataLength', CMQC.MQ_EXIT_DATA_LENGTH, MQLONG_TYPE],
            ['MsgExitsDefined', (0), MQLONG_TYPE],
            ['SendExitsDefined', (0), MQLONG_TYPE],
            ['ReceiveExitsDefined', (0), MQLONG_TYPE],
            ['MsgExitPtr', 0, 'P'],
            ['MsgUserDataPtr', 0, 'P'],
            ['SendExitPtr', 0, 'P'],
            ['SendUserDataPtr', 0, 'P'],
            ['ReceiveExitPtr', 0, 'P'],
            ['ReceiveUserDataPtr', 0, 'P'],
            # Version 4
            ['ClusterPtr', 0, 'P'],
            ['ClustersDefined', (0), MQLONG_TYPE],
            ['NetworkPriority', (0), MQLONG_TYPE],
            ['LongMCAUserIdLength', (0), MQLONG_TYPE],
            ['LongRemoteUserIdLength', (0), MQLONG_TYPE],
            # Version 5
            ['LongMCAUserIdPtr', 0, 'P'],
            ['LongRemoteUserIdPtr', 0, 'P'],
            ['MCASecurityId', b'', '40s'],
            ['RemoteSecurityId', b'', '40s'],
            # Version 6
            ['SSLCipherSpec', b'', '32s'],
            ['SSLPeerNamePtr', 0, 'P'],
            ['SSLPeerNameLength', (0), MQLONG_TYPE],
            ['SSLClientAuth', (0), MQLONG_TYPE],
            ['KeepAliveInterval', -1, MQLONG_TYPE],
            ['LocalAddress', b'', '48s'],
            ['BatchHeartbeat', (0), MQLONG_TYPE],
            # Version 7
            ['HdrCompList', [(0), (-1)], '2' + MQLONG_TYPE],
            ['MsgCompList', [0] + 15 * [(-1)], '16' + MQLONG_TYPE],
            ['CLWLChannelRank', (0), MQLONG_TYPE],
            ['CLWLChannelPriority', (0), MQLONG_TYPE],
            ['CLWLChannelWeight', (50), MQLONG_TYPE],
            ['ChannelMonitoring', (0), MQLONG_TYPE],
            ['ChannelStatistics', (0), MQLONG_TYPE],
            # Version 8
            ['SharingConversations', 10, MQLONG_TYPE],
            ['PropertyControl', 0, MQLONG_TYPE],      # 0 = MQPROP_COMPATIBILITY
            ['MaxInstances', 999999999, MQLONG_TYPE],
            ['MaxInstancesPerClient', 999999999, MQLONG_TYPE],
            ['ClientChannelWeight', 0, MQLONG_TYPE],
            ['ConnectionAffinity', 1, MQLONG_TYPE],  # 1 = MQCAFTY_PREFERRED
            # Version 9
            ['BatchDataLimit', 5000, MQLONG_TYPE],
            ['UseDLQ', 2, MQLONG_TYPE],
            ['DefReconnect', 0, MQLONG_TYPE],
            # Version 10
            ['CertificateLabel', b'', '64s'],
            # Version 11 # This is in 9.1 - can use as base
        ]

        # The MQ 12 additional field is not relevant for client connections
        # but we'll put it here for completeness.
        cd_current_version = ibmmqc.__strucversions__.get("cd", 1)
        if cd_current_version >= CMQXC.MQCD_VERSION_12:
            opts += [['SPLProtection', 0, MQLONG_TYPE]]
            if MQLONG_TYPE == 'i':
                opts += [['__pad', b'', '4s']]

        # In theory, the pad should've been placed right before the 'MsgExitPtr'
        # attribute, however setting it there makes no effect and that's why
        # it's being set here, as a last element in the list.

        super().__init__(tuple(opts), **kw)


# Backward compatibility
cd = CD  # pylint: disable=invalid-name



"""MQTM, MQTMC2: Trigger Monitor structures"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

# ################################################################################################################################

class TM(MQOpts):
    """ Construct an MQTM Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        super().__init__(tuple([
            ['_StrucId', CMQC.MQTM_STRUC_ID, '4s'],
            ['Version', CMQC.MQTM_VERSION_1, MQLONG_TYPE],
            ['QName', b'', '48s'],
            ['ProcessName', b'', '48s'],
            ['TriggerData', b'', '64s'],
            ['ApplType', 0, MQLONG_TYPE],
            ['ApplId', b'', '256s'],
            ['EnvData', b'', '128s'],
            ['UserData', b'', '128s']]), **kw)

# ################################################################################################################################

class TMC2(MQOpts):
    """ Construct an MQTMC2 Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        super().__init__(tuple([
            ['_StrucId', CMQC.MQTMC_STRUC_ID, '4s'],
            ['Version', CMQC.MQTMC_VERSION_2, '4s'],
            ['QName', b'', '48s'],
            ['ProcessName', b'', '48s'],
            ['TriggerData', b'', '64s'],
            ['ApplType', b'', '4s'],
            ['ApplId', b'', '256s'],
            ['EnvData', b'', '128s'],
            ['UserData', b'', '128s'],
            ['QMgrName', b'', '48s']]), **kw)



"""MQBNO: Balancing Options"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class BNO(MQOpts):
    """ Construct an MQBNO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):

        opts = [['_StrucId', CMQC.MQBNO_STRUC_ID, '4s'],
                ['Version', CMQC.MQBNO_VERSION_1, MQLONG_TYPE],
                ['ApplType', CMQC.MQBNO_BALTYPE_SIMPLE, MQLONG_TYPE],
                ['Timeout', CMQC.MQBNO_TIMEOUT_AS_DEFAULT, MQLONG_TYPE],
                ['Options', CMQC.MQBNO_OPTIONS_NONE, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)



"""Setup tests environment."""
import os.path
from unittest import TestCase
from unittest import main  # pylint: disable=unused-import

import ibmmq as mq

import config  # noqa
import utils  # noqa

class Tests(TestCase):
    """Setup and tearsdown tests environment."""

    version = '0000000'

    queue_name = ''
    queue_manager = ''
    channel = ''
    host = ''
    port = ''
    user = ''
    password = ''

    prefix = ''

    qmgr = None  # type: mq.QueueManager
    pcf = None  # type: mq.PCFExecute

    CHCKLOCL = None

    @classmethod
    def setUpClass(cls):
        """Initialize test environment."""
        cls.prefix = os.environ.get('PY_IBMMQ_TEST_OBJECT_PREFIX', 'PYIBMMQ.')

        # max length of queue names is 48 characters
        cls.queue_name = '{prefix}MSG.QUEUE'.format(prefix=config.MQ.QUEUE.PREFIX)
        cls.queue_manager = config.MQ.QM.NAME
        cls.channel = config.MQ.QM.CHANNEL
        cls.app_channel = config.MQ.QM.APP_CHANNEL
        cls.host = config.MQ.QM.HOST
        cls.port = config.MQ.QM.PORT
        cls.user = config.MQ.QM.USER
        cls.password = config.MQ.QM.PASSWORD

        cls.app_user = config.MQ.QM.APP_USER
        cls.app_password = config.MQ.QM.APP_PASSWORD

        cls.conn_info = '{0}({1})'.format(cls.host, cls.port)

        if mq.__mqbuild__ == 'server':
            cls.qmgr = mq.QueueManager(cls.queue_manager)
        else:
            cls.qmgr = mq.QueueManager(None)
            cls.qmgr.connectTCPClient(cls.queue_manager, mq.CD(), cls.channel,
                                      cls.conn_info, cls.user, cls.password)

        cls.pcf = mq.PCFExecute(cls.qmgr, response_wait_interval=30000)

        cls.version = cls.inquire_qmgr_version().decode()

    @classmethod
    def tearDownClass(cls):
        """Clear test environment."""
        if cls.pcf.is_connected:
            cls.pcf.disconnect()
        if cls.qmgr.is_connected:
            cls.qmgr.disconnect()

    def setUp(self):
        """Set up tesing environment."""

    def tearDown(self):
        """Clear test environment."""

    @classmethod
    def inquire_qmgr_version(cls):
        """Inqure Queue Manager version."""
        return cls.qmgr.inquire(mq.CMQC.MQCA_VERSION)

    def create_queue(self, queue_name, max_depth=5000, attrs=None):
        """Create queue."""
        if not attrs:
            attrs = []
            attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_Q_NAME, String=utils.py3str2bytes(queue_name)))
            attrs.append(mq.CFIN(Parameter=mq.CMQC.MQIA_Q_TYPE, Value=mq.CMQC.MQQT_LOCAL))
            attrs.append(mq.CFIN(Parameter=mq.CMQC.MQIA_MAX_Q_DEPTH, Value=max_depth))
            attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_REPLACE, Value=mq.CMQCFC.MQRP_YES))

        self.pcf.MQCMD_CREATE_Q(attrs)

    def delete_queue(self, queue_name):
        """Delete queue."""
        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_Q_NAME, String=utils.py3str2bytes(queue_name)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_PURGE, Value=mq.CMQCFC.MQPO_YES))
        self.pcf.MQCMD_DELETE_Q(attrs)

    def create_channel(self, channel_name, attrs=None):
        """Create channle."""
        if not attrs:
            attrs = []
            attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME, String=utils.py3str2bytes(channel_name)))
            attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACH_CHANNEL_TYPE, Value=mq.CMQXC.MQCHT_SVRCONN))
            attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_REPLACE, Value=mq.CMQCFC.MQRP_YES))
        self.pcf.MQCMD_CREATE_CHANNEL(attrs)

    def delete_channel(self, channel_name):
        """Delete channel."""
        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME, String=utils.py3str2bytes(channel_name)))
        self.pcf.MQCMD_DELETE_CHANNEL(attrs)

    def create_auth_rec(self, attrs):
        """Create authentication recoed."""
        self.pcf.MQCMD_SET_CHLAUTH_REC(attrs)

    def delete_auth_rec(self, attrs):
        """Delete authentication recoed."""
        self.pcf.MQCMD_SET_CHLAUTH_REC(attrs)

    @classmethod
    def edit_qmgr(cls, attrs):
        """Edit connected Queue Manager."""
        cls.pcf.MQCMD_CHANGE_Q_MGR(attrs)



"""MQMD: Message Descriptor"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class MD(MQOpts):
    """ Construct an MQMD Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    This is initialised as a V2 structure as every in-support version of MQ supports
    that level.
    """
    def __init__(self, **kw):
        super().__init__(tuple([
            ['_StrucId', CMQC.MQMD_STRUC_ID, '4s'],
            ['Version', CMQC.MQMD_VERSION_2, MQLONG_TYPE],
            ['Report', CMQC.MQRO_NONE, MQLONG_TYPE],
            ['MsgType', CMQC.MQMT_DATAGRAM, MQLONG_TYPE],
            ['Expiry', CMQC.MQEI_UNLIMITED, MQLONG_TYPE],
            ['Feedback', CMQC.MQFB_NONE, MQLONG_TYPE],
            ['Encoding', CMQC.MQENC_NATIVE, MQLONG_TYPE],
            ['CodedCharSetId', CMQC.MQCCSI_Q_MGR, MQLONG_TYPE],
            ['Format', b'', '8s'],
            ['Priority', CMQC.MQPRI_PRIORITY_AS_Q_DEF, MQLONG_TYPE],
            ['Persistence', CMQC.MQPER_PERSISTENCE_AS_Q_DEF, MQLONG_TYPE],
            ['MsgId', b'', '24s'],
            ['CorrelId', b'', '24s'],
            ['BackoutCount', 0, MQLONG_TYPE],
            ['ReplyToQ', b'', '48s'],
            ['ReplyToQMgr', b'', '48s'],
            ['UserIdentifier', b'', '12s'],
            ['AccountingToken', b'', '32s'],
            ['ApplIdentityData', b'', '32s'],
            ['PutApplType', CMQC.MQAT_NO_CONTEXT, MQLONG_TYPE],
            ['PutApplName', b'', '28s'],
            ['PutDate', b'', '8s'],
            ['PutTime', b'', '8s'],
            ['ApplOriginData', b'', '4s'],
            ['GroupId', b'', '24s'],
            ['MsgSeqNumber', 1, MQLONG_TYPE],
            ['Offset', 0, MQLONG_TYPE],
            ['MsgFlags', CMQC.MQMF_NONE, MQLONG_TYPE],
            ['OriginalLength', CMQC.MQOL_UNDEFINED, MQLONG_TYPE]]), **kw)


# Backward compatibility
md = MD  # pylint: disable=invalid-name



"""Some basic functions mostly needed for data conversions"""

# Copyright (c) 2025, 2026 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

import struct
from typing import Callable, Union

class EncodingDefault:
    """Global encoding options"""
    ccsid = 1208
    bytes_encoding = 'utf8'

    # How are unicode strings converted to bytes for MQI fields? By default we pick an ASCII encoding, as
    # that is good for the majority of codepages where an MQ app is running.
    # A z/OS system might need to override this to ebcdic: "cp500" perhaps.
    mqi_encoding = 'ascii'


class OTelOptions:
    """This public class allows applications to set OTel-related options when processing messages"""
    def __init__(self, **kwargs):
        self.remove_rfh2 = kwargs['remove_rfh2'] if 'remove_rfh2' in kwargs else False

class OTelFunctions:
    """This internal class contains indirections to the OTel tracing functions. They
    are only set to real values when we are in an OTel-enabled environment
    """
    disc = None  # type: Union[None,Callable]
    open = None  # type: Union[None,Callable]
    close = None  # type: Union[None,Callable]
    put_trace_before = None  # type: Union[None,Callable]
    put_trace_after = None  # type: Union[None,Callable]
    get_trace_before = None  # type: Union[None,Callable]
    get_trace_after = None  # type: Union[None,Callable]

def padded_count(count: int, boundary: int = 4) -> int:
    """Calculate padded bytes count
    """
    return count + ((boundary - count & (boundary - 1)) & (boundary - 1))

def is_unicode(s) -> bool:
    """ Returns True if input arg is a Python 3 string (aka Python 2 unicode). False otherwise.
    """
    if isinstance(s, str) and not isinstance(s, bytes):
        return True
    return False

def ensure_not_unicode(value) -> None:
    """While we will mostly convert Unicode strings to bytes for MQI fields, there are a few remaining
    places where we don't want to do that - primarily in message body contents. Make sure the message
    data is what the application expects, by forcing them to use bytes.
    """
    if is_unicode(value):
        msg = 'Python 3 style string (unicode) found but not allowed here: `{0}`. Convert to bytes.'
        raise TypeError(msg.format(value))

def ensure_strings_are_bytes(s, encoding=EncodingDefault.mqi_encoding) -> bytes:
    """MQI CHAR fields need to be handled as byte arrays, even if provided
    via Python 3 unicode strings
    """
    if is_unicode(s):
        return s.encode(encoding)
    return s


# For compatibility
ensure_bytes = ensure_strings_are_bytes

#
# 64bit suppport courtesy of Brent S. Elmer, Ph.D. (mailto:webe3vt@aim.com)
#
# On 64 bit machines when MQ is compiled 64bit, MQLONG is an int defined
# in /opt/mqm/inc/cmqc.h or wherever your MQ installs to.
#
# On 32 bit machines, MQLONG is a long and many other MQ data types are set to MQLONG
#
# So, set MQLONG_TYPE to 'i' for 64bit MQ and 'l' for 32bit MQ so that the
# conversion from the Python data types to C data types in the MQ structures
# will work correctly.
#
# However, note that V2 of this package no longer supports running on 32-bit systems anyway
# so we're going to throw an error.

# Are we running 64 bit?
if struct.calcsize('P') == 8:
    MQLONG_TYPE = 'i'  # 64 bit
else:
    MQLONG_TYPE = 'l'  # 32 bit
    raise SystemError("32-bit systems are no longer supported in this package")

INTEGER64_TYPE = 'q'

# A debug tool
def _dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example demonstrates setting and viewing message properties. It assumes
the queue is empty before starting. The value and class associated with the property
are logged.
"""

import logging

import ibmmq as mq

def read_props_wildcard():
    """A function to iterate through the properties, as it's used twice in this example"""
    props_to_read = True
    opts = mq.CMQC.MQIMPO_INQ_FIRST
    while props_to_read:
        try:
            property_value, prop_name = get_msg_h.properties.get('usr.%', impo_options=opts)
            logging.info('  Property name: `%s`, property value: `%s` type %s', prop_name, property_value, type(property_value))
            opts = mq.CMQC.MQIMPO_INQ_NEXT

        except mq.MQMIError as e:
            if e.reason == mq.CMQC.MQRC_PROPERTY_NOT_AVAILABLE:
                logging.info('No more properties')
                props_to_read = False
            else:
                raise e


logging.basicConfig(level=logging.INFO)

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'

conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'
logging.info('Starting message property example.')

# Connect to the qmgr
qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

# Create a message handle to manipulate properties
put_msg_h = mq.MessageHandle(qmgr)

# Define the property names we are going to use
property_base_name = 'Property_'

p = []
p += [property_base_name + '0']
p += [property_base_name + '1']
p += [property_base_name + '2']

# Set values of different types
put_msg_h.properties.set(p[0], 'MyStringPropVal')  # Default type is CMQC.MQTYPE_STRING
put_msg_h.properties.set(p[1], 42, property_type=mq.CMQC.MQTYPE_INT32)
put_msg_h.properties.set(p[2], 3.13, property_type=mq.CMQC.MQTYPE_FLOAT64)


pmo = mq.PMO(Version=mq.CMQC.MQPMO_VERSION_3)  # PMO v3 is required if you want to use message handles
pmo.OriginalMsgHandle = put_msg_h.msg_handle
opts = mq.CMQC.MQOO_OUTPUT
put_md = mq.MD(Version=mq.CMQC.MQMD_CURRENT_VERSION)

put_queue = mq.Queue(qmgr, queue_name, opts)
put_queue.put(b'Property testing', put_md, pmo)

# Now retrieve the message and inspect its properties
get_msg_h = mq.MessageHandle(qmgr)

gmo = mq.GMO(Version=mq.CMQC.MQGMO_CURRENT_VERSION)
gmo.Options = mq.CMQC.MQGMO_PROPERTIES_IN_HANDLE + mq.CMQC.MQGMO_SYNCPOINT
gmo.MsgHandle = get_msg_h.msg_handle

get_md = mq.MD()
get_queue = mq.Queue(qmgr, queue_name)
message_body = get_queue.get(None, get_md, gmo)

# Two different ways to iterate through the query
# In the first, we don't know the property name so use a wildcard
logging.info('')
logging.info('Message received')
logging.info('Using wildcards to retrieve properties:')
read_props_wildcard()

# In the second, we ask for an explicit property
logging.info('')
logging.info('Using explicit names to retrieve properties:')

for prop in p:
    property_value = get_msg_h.properties.get(prop)
    logging.info('  Property name: `%s`, property value: `%s` type %s', prop, property_value, type(property_value))

# Delete one of the properties
logging.info('')
logging.info("Deleting property: '%s'", p[1])
get_msg_h.properties.dlt(p[1])

# And see if the delete was successful
logging.info('')
logging.info('Using wildcards to retrieve properties again:')
read_props_wildcard()

# Cleanup and exit
put_msg_h.dlt()
get_msg_h.dlt()

put_queue.close()
get_queue.close()

# The message was read in syncpoint, so we need to commit the removal
qmgr.commit()
qmgr.disconnect()



"""
A common interface to logging functions that can be used by the ibmmq package.
All log messages are shown as coming from the same module - we don't have a per-file
logger.

It's expected that the strings passed here are already formatted
for simplicity (ie use f"xxx{var}" in the caller). Though you probably could use the %-style
with additional parameters to the functions as an alternative.
"""

# Copyright (c) 2025, 2026 IBM Corporation and other Contributors. All Rights Reserved.

import os
import logging
from ibmmq import ibmmqc

# We do NOT use basicConfig as that could affect any application logging; instead,
# set the configuration specific to this module's logger.
logger = logging.getLogger('ibmmq')
trace_level = False

def trace(msg, *args):
    """The Python logger doesn't have a separate trace level
    so we fake it. Trace is more detailed than debug.
    """
    if logger and trace_level:
        logger.debug(msg, *args)

def debug(msg, *args):
    """Add some indent to the message if we are also
    using the TRACE pseudo-loglevel
    """
    indent = ""
    if trace_level:
        indent = "  "
    if logger:
        logger.debug(indent + msg, *args)

def critical(msg, *args):
    """Record a critical message. Not used in this package, but defined
    here for completeness.
    """
    if logger:
        logger.critical(msg, *args)

def error(msg, *args):
    """Record an error message"""
    if logger:
        logger.error(msg, *args)

def warning(msg, *args):
    """Record a warning message"""
    if logger:
        logger.warning(msg, *args)

def trace_entry(s, *args):
    """Record entry to a function"""
    trace("> " + s, *args)

def trace_exit(s: str, **kwargs):
    """Record exit of a function. The "ep" attribute lets us
    annotate when there are multiple potential exits from a function (usually
    because of error handling). For a given function, the exitpoint value
    should be unique. There may also be an MQRC value if the underlying MQI
    function has returned a non-zero MQCC.
    """

    expt = f" EXPT:{kwargs['ep']}" if 'ep' in kwargs else ""
    mqrc = f" MQRC:{kwargs['rc']}" if 'rc' in kwargs else ""
    trace("< " + s + expt + mqrc)


# Create a logger for Python.
# Also configure the C layer if we've asked for trace/debug output and
# once there's actually some debug/trace points in there. For now, this
# is disabled.
enable_native_logging = False

# TRACE takes precedence over DEBUG
if os.environ.get('MQIPY_TRACE'):
    # There's no separate "TRACE" level in the Python logging API so we separate
    # it from DEBUG with a local flag.
    level = logging.DEBUG
    # enable_native_logging = True
    trace_level = True
elif os.environ.get('MQIPY_DEBUG'):
    level = logging.DEBUG
    # enable_native_logging = True
else:
    level = logging.WARN  # We only use WARN/ERROR in normal use, so no point in enabling INFO

logger.setLevel(level)

# If you've asked for a specfic filename, then override any default
# logging propagation and only allow that file to be used. If the
# file cannot be created/written, then there is an exception generated.
filename = os.environ.get('MQIPY_LOG_FILENAME')
if filename:
    for h in logger.handlers:
        logger.removeHandler(h)
    if filename == "stderr":
        h = logging.StreamHandler()
    else:
        h = logging.FileHandler(filename)  # Default constructor has mode=append

    # consider adding: "%(threadName)s %(thread)d"
    h.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s %(name)-8s %(message)s', datefmt='%H:%M:%S'))
    logger.addHandler(h)
    logger.propagate = False
    if enable_native_logging:
        ibmmqc.MQLOGCF(level, filename)
else:
    logger.propagate = True
    p = logger
    found_handler = False
    while p.parent:
        p = p.parent
        if len(p.handlers) != 0:
            # print(f"Found a handler in logger {p}. List={p.handlers}")
            found_handler = True
            break
        # else:
        #    print(f"No handlers found for {p}")

    # If there is no available handler, then attach a stderr stream to our logger
    if not found_handler:
        # print(f"Adding stream handler to {logger}")
        logger.propagate = False
        logger.addHandler(logging.StreamHandler())

    if enable_native_logging:
        ibmmqc.MQLOGCF(level)



"""MQOpts: Base class for all MQI structures"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

import ctypes
import struct

from mqcommon import *
from mqerrors import *
from ibmmq import CMQC

try:
    from typing import Any, Optional, Union, Dict
except ImportError:
    pass


# These are fields from the MQI structures that are known to always be binary
# and which should not be converted back to strings. There are not many, so it's
# reasonable to list them here. But any new MQI fields will all need to be thought
# about to see if they need to be added to the list.
_binary_fields = ["AccountingToken",
                  "AlternateSecurityId",
                  "ConnectionId",
                  "ConnTag",
                  "CorrelId",
                  "Facility",  # CIH
                  "GroupId",
                  "MCASecurityId",
                  "MsgId",
                  "MsgToken",
                  "PubAccountingToken",
                  "RemoteSecurityId",
                  "SubCorrelId",
                  "TranInstanceId"  # IIH
                  ]

# ################################################################################################################################
# MQI Python<->C Structure mapping. MQI uses lots of parameter passing
# structures in its API. These classes are used to set/get the
# parameters in the style of Python dictionaries & keywords. Pack &
# unpack calls are used to convert between python class attributes and
# 'C' structures, suitable for passing in/out of C functions.
#
# The MQOpts class defines common operations for structure definition,
# default values setting, member set/get and translation to & from 'C'
# structures. Specializations construct MQOpts with a list specifying
# structure member names, their default values, and pack/unpack
# formats. MQOpts uses this list to setup class attributes
# corresponding to the structure names, set up attribute defaults, and
# builds a format string usable by the struct package to translate to
# 'C' structures.

class MQOpts:
    """ Base class for packing/unpacking MQI Option structures. It is
    constructed with a list defining the member/attribute name,
    default value (from the CMQC module) and the member pack format
    (see the struct module for the formats). The list format is:

      [['Member0', CMQC.DEFAULT_VAL0, 'fmt1']
       ['Member1', CMQC.DEFAULT_VAL1, 'fmt2']
         ...
      ]

    MQOpts defines common methods to allow structure members to be
    set/get as attributes (foo.Member0 = 42), set/get as dictionary
    items (foo['Member0'] = 42) or set as keywords (foo.set(Member0 =
    42, Member1 = 'flipperhat'). The ctor can be passed an optional
    keyword list to initialize the structure members to non-default
    values. The get method returns all attributes as a dictionary.

    The pack() method packs all members into a 'C' structure according
    to the format specifiers passed to the ctor. The packing order is
    as specified in the list passed to the ctor. Pack returns a string
    buffer, which can be passed directly to the MQI 'C' calls.

    For packing, strings are truncated or padded with null bytes as appropriate to
    make them fit the given field length. There is no warning of overlong strings
    such as trying to use a 64-char queue name. For unpacking, the resulting bytes object
    always has exactly the specified number of bytes.

    The unpack() method does the opposite of pack. It unpacks a
    buffer into an MQOpts instance.

    Applications are not expected to use MQOpts directly. Instead,
    MQOpts is sub-classed as particular MQI structures.
    """

    def __init__(self, memlist, **kw):
        # type: (Union[list,tuple], Any) -> None
        """ Initialise the option structure. 'list' is a list of structure
        member names, default values and pack/unpack formats. 'kw' is an
        optional keyword dictionary that may be used to override default
        values set by MQOpts sub-classes.
        """

        self.__list = memlist[:]
        self.__format = ''

        # Dict to store c_char arrays to prevent memory addresses
        # from getting overwritten
        self.__vs_ctype_store = {}  # type: Dict[str, Any]

        # Create the structure members as instance attributes and build
        # the struct.pack/unpack format string. The attribute name is
        # identical to the 'C' structure member name, except for some *Ptr fields.
        for i in memlist:
            setattr(self, i[0], i[1])
            try:
                i[3]
            except LookupError:
                i.append(1)
            self.__format = self.__format + i[2] * i[3]
        self.set(**kw)

    def pack(self) -> bytes:
        """ Pack the attributes into a 'C' structure to be passed to MQI
        calls. The pack order is as defined to the MQOpts
        ctor. Returns the structure as a bytes buffer.

        This may also be useful when you need to construct message bodies that have
        chained MQ headers (eg an MQDLH followed by the real body). For example,
           q.put(dlh.pack() + bytes(text, 'utf8'), md)
        """

        # Build tuple for struct.pack() argument. Start with format string.
        args = [self.__format]  # type: list[Any]

        # Now add the current attribute values to the tuple
        # In most cases, strings can be automatically converted from Unicode to the byte
        # array needed in an MQCHAR buffer. There are a few fields, however, where we really
        # do want to enforce that it's bytes for the MQBYTE[] buffers. None of these fields
        # are in the list elements.
        for i in self.__list:
            v = getattr(self, i[0])

            # Flatten attribs that are arrays
            if isinstance(v, list):
                for x in v:
                    args.append(ensure_strings_are_bytes(x))
            else:
                # print(f"Field: {i[0]} format '{i[2]}' is of value {v}: type {type(v)}")
                if i[0] in _binary_fields:
                    if not isinstance(v, bytes):
                        err = f'{i[0]} must be a byte array'
                        raise TypeError(err)
                elif v is None:
                    # If a string field is set to None, the pack() function
                    # fails with an unobvious error. Let's try to diagnose
                    # it a bit earlier. And if there's a Pointer that has escaped being
                    # set to NULL, then fix it up.
                    if i[2].endswith('s'):
                        err = f'Class:{type(self).__name__} Field:{i[0]} must not be None'
                        raise TypeError(err)
                    if i[2] == 'P':
                        v = 0

                args.append(ensure_strings_are_bytes(v))

        # print(f"Args to be packed are {args}")

        return struct.pack(*args)

    def unpack(self, buff: bytes):
        """ Unpack a 'C' structure 'buff' into self. Also returns self.
        """
        ensure_not_unicode(buff)

        # Increase buff length to the current MQOpts structure size
        diff_length = self.get_length() - len(buff)
        if diff_length > 0:
            buff += b'\x00' * diff_length

        # Unpack returns a tuple of the unpacked data, in the same
        # order (I hope!) as in the constructor's list arg.
        r = struct.unpack(self.__format, buff)
        x = 0
        for i in self.__list:

            if isinstance(i[1], list):
                ll = []
                for _j in range(i[3]):
                    ensure_not_unicode(r[x])
                    ll.append(r[x])
                    x = x + 1
                setattr(self, i[0], ll)
            else:
                ensure_not_unicode(r[x])  # Python 3 bytes check
                setattr(self, i[0], r[x])
                x = x + 1
        return self

    def to_string(self, encoding=EncodingDefault.bytes_encoding):
        """Given an MQI class (eg MQDLH), try to convert any string-like byte arrays into Python 3 strings.
        Fields that are known to be truly binary (like MsgId) are excluded from the conversion, but there
        may still be times that the output is not what you expected. This does modify the input structure
        so you might want to use a copy if you expect to continue to want to work with the original binary elements.
        """
        for i in self.__list:
            k = i[0]        # The attribute name
            v = self[i[0]]  # The attribute value
            # There are no regular MQI structure fields that contain lists of strings/byte arrays:
            # the PCF MQCFSL is handled separately.
            if isinstance(v, bytes):
                if k not in _binary_fields:
                    try:
                        null_index = v.find(0)
                        if null_index != -1:
                            v = v[:null_index]
                        setattr(self, k, v.decode(encoding).strip())
                    except UnicodeError:
                        pass
                else:
                    pass
                    # print("Not decoding bytes for:",k)
            else:
                pass
                # print("Not decoding ", k)
        return self

    def set(self, **kw: Dict[str, Any]):
        """ Set a structure member using the keyword dictionary 'kw'.
        An AttributeError exception is raised for invalid member names.
        """

        for k, v in kw.items():
            # Only set if the attribute already exists. getattr raises
            # an exception if it doesn't.
            getattr(self, str(k))
            setattr(self, str(k), ensure_strings_are_bytes(v))

    def __setitem__(self, key: str, value: Any) -> None:
        """ Set the structure member attribute 'key' to 'value', as in obj['Attr'] = 42.
        """
        # Only set if the attribute already exists. getattr raises an
        # exception if it doesn't.
        getattr(self, key)
        setattr(self, key, ensure_strings_are_bytes(value))

    def get(self):
        # type: () -> dict
        """ Return a dictionary of the current structure member values. The dictionary is keyed by a 'C' member name.
        """
        d = {}
        for i in self.__list:
            d[i[0]] = getattr(self, i[0])
        return d

    def __getitem__(self, key: str) -> Any:
        """Return the member value associated with key, as in print obj['Attr'].
        """
        return getattr(self, key)

    def _remove(self, key: str) -> None:
        """Remove an attribute from the object. Not meant for general use.
        This removes both the attribute and knowledge about how it is formatted
        """
        # Have to convert tuples into a list so it can be modified
        tmp_list = list(self.__list)

        for item in tmp_list:
            if item[0] == key:
                tmp_list.remove(item)
                try:
                    delattr(self, key)
                except AttributeError:
                    pass
                break

        # And then convert back to the tuple format
        self.__list = tuple(tmp_list)

    def __str__(self) -> str:
        rv = ''
        for i in self.__list:
            rv = rv + str(i[0]) + ': ' + str(getattr(self, i[0])) + '\n'
        # Chop the trailing newline
        return rv[:-1]

    def __repr__(self) -> str:
        """ Return the packed buffer as a printable string.
        """
        return str(self.pack())

    def get_length(self) -> int:
        """ Returns the length of the (would be) packed buffer.
        """
        return struct.calcsize(self.__format)

    def set_vs(self, vs_name: str, vs_value: Union[bytes, str, None] = None, vs_offset: int = 0, vs_buffer_size: int = 0, vs_ccsid: int = 0) -> None:
        """ This method aids in the setting of the MQCHARV (variable length
        string) types in MQ structures. The type contains a pointer to a
        variable length string. A common example of a MQCHARV type
        is the ObjectString in the MQOD structure.
        In this module the ObjectString is defined as 5 separate
        elements (as per MQCHARV):
        ObjectStringVSPtr - Pointer
        ObjectStringVSOffset - Long
        ObjectStringVSBufSize - Long
        ObjectStringVSLength - Long
        ObjectStringVSCCSID - Long
        """

        vs_value = ensure_strings_are_bytes(vs_value)  # allow known args be a string in Py3

        # if the VSPtr name is passed - remove VSPtr to be left with name.
        if vs_name.endswith('VSPtr'):
            vs_name_vsptr = vs_name
        else:
            vs_name_vsptr = vs_name + 'VSPtr'

        vs_name_vsoffset = vs_name + 'VSOffset'
        vs_name_vsbuffsize = vs_name + 'VSBufSize'
        vs_name_vslength = vs_name + 'VSLength'
        vs_name_vsccsid = vs_name + 'VSCCSID'

        c_vs_value = None
        c_vs_value_p = 0  # type: Optional[int]

        if vs_value is not None:
            c_vs_value = ctypes.create_string_buffer(vs_value)
            c_vs_value_p = ctypes.cast(c_vs_value, ctypes.c_void_p).value

        self[vs_name_vsptr] = c_vs_value_p
        self[vs_name_vsoffset] = vs_offset
        self[vs_name_vsbuffsize] = vs_buffer_size
        self[vs_name_vslength] = len(vs_value)
        self[vs_name_vsccsid] = vs_ccsid

        # Store c_char array object so memory location does not get overwritten
        self.__vs_ctype_store[vs_name] = c_vs_value

    def get_vs(self, vs_name):
        # type: (str) -> Union[bytes, str, None]
        """ This method returns the string to which the VSPtr pointer points to.
        """
        # if the VSPtr name is passed - remove VSPtr to be left with name.
        if vs_name.endswith('VSPtr'):
            vs_name_vsptr = vs_name
        else:
            vs_name_vsptr = vs_name + 'VSPtr'

        c_vs_value = None
        c_vs_value_p = self[vs_name_vsptr]
        if c_vs_value_p != 0:
            c_vs_value = ctypes.cast(c_vs_value_p, ctypes.c_char_p).value

        return c_vs_value

    # The C MQI will often define 3 fields to refer to a string:
    #   xxPtr, xxLength, xxOffset
    # To make it easier to use these in other languages, we expose these
    # via a single "xx" field and then fill in the other 2 pieces automatically.
    # The class has "xx", "_xxLength" and "_xxOffset" fields to guide which fields
    # are for internal use. But they are needed to maintain the same structure layout.
    # Offset is always 0. Length is set to the length of the string/byte array excluding
    # trailing NUL.

    def _set_ptr_field(self, field: str, value: Any):
        field_name_ptr = field

        # SSLPeerNamePtr has only ptr/len and we shouldn't change the name to match the
        # style of later similar attributes for compatibility reasons
        if field.endswith('Ptr'):
            field_base = field.replace('Ptr', '')
        else:
            field_base = field

        field_name_offset = '_' + field_base + 'Offset'
        field_name_length = '_' + field_base + 'Length'

        # This error could happen if the package has been built against
        # a lower level of MQ than the field requires. Look for the hidden
        # attribute ("_xxLength") first. There may be backward compatibility
        # requirements on allowing the unhidden ("xxLength") version. If neither
        # are available, then raise the error.
        if not hasattr(self, field_name_length):
            field_name_length = field_base + 'Length'
            if not hasattr(self, field_name_length):
                raise MQMIError(CMQC.MQCC_FAILED, CMQC.MQRC_WRONG_VERSION)

        if hasattr(self, field_name_offset):
            self[field_name_offset] = 0

        self[field_name_length] = 0

        value = ensure_strings_are_bytes(value)
        c_value = None
        c_value_p = 0  # type: Optional[int]
        if value is not None and value != 0:
            c_value = ctypes.create_string_buffer(value)
            c_value_p = ctypes.cast(c_value, ctypes.c_void_p).value
            self[field_name_length] = len(c_value) - 1  # Ignore the trailing NUL that Python adds

        self[field_name_ptr] = c_value_p

        # Store c_char array object so memory location does not get overwritten
        self.__vs_ctype_store[field] = c_value



"""Test connection with TLS(SSL)."""

import os.path
import unittest

import ibmmq as mq

from test_setup import Tests
from test_setup import main
import utils

# pylint: disable=pointless-string-statement
"""
Preparation:
    Define env variable:
        PY_IBMMQ_TEST_TLS_SKIP = 0

Create keydb files:
    Default keydb localtion is:
        Server:
            Unix: /var/mqm/qmgrs/queue_manager_name/ssl
        Client:
            Unix: user HOME directory

Command for keydb creation (for Unix):

runmqakm -keydb -create -db /var/mqm/qmgrs/MQTEST/ssl/mqtest.kdb -pw Secret13 -stash
runmqakm -cert -create -db /var/mqm/qmgrs/MQTEST/ssl/mqtest.kdb -type kdb -pw Secret13 \
    -label mqtest -dn CN=mqtest -size 2048 -x509version 3 -expire 365 -fips -sig_alg SHA256WithRSA

runmqakm -keydb -create -db ~/client.kdb -pw Secret13 -stash
runmqakm -cert -create -db ~/client.kdb -type kdb -pw Secret13 \
    -label client -dn CN=client -size 2048 -x509version 3 -expire 365 -fips -sig_alg SHA256WithRSA

runmqakm -cert -extract -db /var/mqm/qmgrs/MQTEST/ssl/mqtest.kdb -pw Secret13 \
    -label mqtest -target ~/mqtest.pem
runmqakm -cert -add -db ~/client.kdb -pw Secret13 \
    -label mqtest -file ~/mqtest.pem

runmqakm -cert -extract -db ~/client.kdb -pw Secret13 \
    -label client -target ~/client.pem
runmqakm -cert -add -db /var/mqm/qmgrs/MQTEST/ssl/mqtest.kdb -pw Secret13 \
    -label client -file ~/client.pem
"""

class TestTLS(Tests):
    """Test Queue.get() method."""

    tls_channel_name = 'SSL.SVRCONN'
    cypher_spec = 'TLS_RSA_WITH_AES_256_CBC_SHA256'
    client_dn = 'CN=client'
    certificate_label_qmgr = 'mqtest'
    certificate_label_client = 'client'
    key_repo_location_client = os.path.expanduser('~')
    key_repo_location_qmgr = '/var/mqm/qmgrs/{}/ssl'

    @classmethod
    def setUpClass(cls):
        """Initialize test environment."""
        cls.skip = int(os.environ.get(
            'PY_IBMMQ_TEST_TLS_SKIP', 1))

        if cls.skip:
            raise unittest.SkipTest('PY_IBMMQ_TEST_TLS_SKIP initialized')

        super(TestTLS, cls).setUpClass()
        cls.tls_channel_name = os.environ.get(
            'PY_IBMMQ_TEST_TLS_CHL_NAME',
            cls.prefix + cls.tls_channel_name).encode()
        cls.cypher_spec = os.environ.get(
            'PY_IBMMQ_TEST_TLS_CYPHER_SPEC',
            cls.cypher_spec).encode()
        cls.client_dn = os.environ.get(
            'PY_IBMMQ_TEST_TLS_CLIENT_DN',
            cls.client_dn).encode()
        cls.certificate_label_qmgr = os.environ.get(
            'PY_IBMMQ_TEST_TLS_CERT_LABEL_QMGR',
            cls.certificate_label_qmgr).encode()
        cls.certificate_label_client = os.environ.get(
            'PY_IBMMQ_TEST_TLS_CERT_LABEL_CLIENT',
            cls.certificate_label_client).encode()
        cls.key_repo_location_qmgr = os.environ.get(
            'PY_IBMMQ_TEST_TLS_KEY_REPO_LOCATION_QMGR',
            cls.key_repo_location_qmgr.format(cls.queue_manager)).encode()
        cls.key_repo_location_client = os.environ.get(
            'PY_IBMMQ_TEST_TLS_KEY_REPO_LOCATION_CLIENT',
            cls.key_repo_location_client).encode()

        cls.key_repo_location_client_path = os.path.join(
            cls.key_repo_location_client,
            cls.certificate_label_client)

        cls.key_repo_location_qmgr_path = os.path.join(
            cls.key_repo_location_qmgr,
            cls.certificate_label_qmgr)

        cls._key_repo_location_qmgr = cls.qmgr.inquire(mq.CMQC.MQCA_SSL_KEY_REPOSITORY)
        cls._certificate_label_qmgr = cls.qmgr.inquire(mq.CMQC.MQCA_CERT_LABEL)

        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_SSL_KEY_REPOSITORY,
                                String=cls.key_repo_location_qmgr_path))
        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_CERT_LABEL,
                                String=cls.certificate_label_qmgr))

        cls.edit_qmgr(attrs)

    @classmethod
    def tearDownClass(cls):
        """Clear test environment."""
        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_SSL_KEY_REPOSITORY,
                                String=cls._key_repo_location_qmgr))
        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_CERT_LABEL,
                                String=cls._certificate_label_qmgr))

        cls.edit_qmgr(attrs)
        super(TestTLS, cls).tearDownClass()

    def setUp(self):
        """Initialize test environment."""
        super().setUp()

        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME,
                                String=utils.py3str2bytes(self.tls_channel_name)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACH_CHANNEL_TYPE,
                                Value=mq.CMQXC.MQCHT_SVRCONN))
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_SSL_CIPHER_SPEC,
                                String=self.cypher_spec))
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_SSL_PEER_NAME,
                                String=self.client_dn))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACH_SSL_CLIENT_AUTH,
                                Value=mq.CMQXC.MQSCA_OPTIONAL))
        attrs.append(mq.CFST(Parameter=mq.CMQC.MQCA_CERT_LABEL,
                                String=self.certificate_label_qmgr))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_REPLACE,
                                Value=mq.CMQCFC.MQRP_YES))
        self.create_channel(self.tls_channel_name, attrs)

        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME,
                                String=utils.py3str2bytes(self.tls_channel_name)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_CHLAUTH_TYPE,
                                Value=mq.CMQCFC.MQCAUT_USERMAP))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_ACTION,
                                Value=mq.CMQCFC.MQACT_REPLACE))
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CLIENT_USER_ID,
                                String=utils.py3str2bytes(self.user)))
        attrs.append(mq.CFIN(Parameter=mq.CMQC.MQIA_CHECK_CLIENT_BINDING,
                                Value=mq.CMQCFC.MQCHK_REQUIRED_ADMIN))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACH_USER_SOURCE,
                                Value=mq.CMQC.MQUSRC_MAP))
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_MCA_USER_ID,
                                String=b'mqm'))

        self.create_auth_rec(attrs)

        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME,
                                String=utils.py3str2bytes(self.tls_channel_name)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_CHLAUTH_TYPE,
                                Value=mq.CMQCFC.MQCAUT_BLOCKUSER))
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_MCA_USER_ID_LIST,
                                String=b'nobody'))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACH_WARNING,
                                Value=mq.CMQC.MQWARN_NO))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_ACTION,
                                Value=mq.CMQCFC.MQACT_REPLACE))
        self.create_auth_rec(attrs)

    def tearDown(self):
        """Delete created objects."""
        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME,
                                String=utils.py3str2bytes(self.tls_channel_name)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_CHLAUTH_TYPE,
                                Value=mq.CMQCFC.MQCAUT_USERMAP))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_ACTION,
                                Value=mq.CMQCFC.MQACT_REMOVEALL))
        self.delete_auth_rec(attrs)

        attrs = []
        attrs.append(mq.CFST(Parameter=mq.CMQCFC.MQCACH_CHANNEL_NAME,
                                String=utils.py3str2bytes(self.tls_channel_name)))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_CHLAUTH_TYPE,
                                Value=mq.CMQCFC.MQCAUT_BLOCKUSER))
        attrs.append(mq.CFIN(Parameter=mq.CMQCFC.MQIACF_ACTION,
                                Value=mq.CMQCFC.MQACT_REMOVEALL))
        self.delete_auth_rec(attrs)

        self.delete_channel(self.tls_channel_name)

        super().tearDown()

    ###########################################################################
    #
    # Real Tests start here
    #
    ###########################################################################

    def test_connection_with_tls(self):
        """Test connection to QueueManager with TLS."""
        qmgr = mq.QueueManager(None)
        conn_info = mq.ensure_bytes('{0}({1})'.format(self.host, self.port))

        cd = mq.CD(Version=mq.CMQXC.MQCD_VERSION_7,  # pylint: disable=C0103
                   ChannelName=self.tls_channel_name,
                   ConnectionName=conn_info,
                   SSLCipherSpec=self.cypher_spec)

        sco = mq.SCO(Version=mq.CMQC.MQSCO_VERSION_5,
                     KeyRepository=os.path.join(self.key_repo_location_client,
                                                self.certificate_label_client),
                     CertificateLabel=self.certificate_label_client)

        opts = mq.CMQC.MQCNO_HANDLE_SHARE_NO_BLOCK

        qmgr.connectWithOptions(self.queue_manager, cd, sco, opts=opts,
                                user=self.user, password=self.password)
        is_connected = qmgr.is_connected
        if is_connected:
            qmgr.disconnect()

        self.assertTrue(is_connected)


if __name__ == '__main__':
    main(module='test_tls')



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

""""
This example shows a way to connect as a client to a remote queue manager, passing
credentials as explicit positional parameters to the connect() method.
"""
import sys
import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)
print(f"Connection succeeded using Python {sys.version_info.major}.{sys.version_info.minor}")

qmgr.disconnect()



"""MQCBD: Callback Descriptor"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class CBD(MQOpts):
    """ Construct an MQCBD Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQCBD_STRUC_ID, '4s'],
                ['Version', CMQC.MQCBD_VERSION_1, MQLONG_TYPE],
                ['CallbackType', CMQC.MQCBT_MESSAGE_CONSUMER, MQLONG_TYPE],
                ['Options', CMQC.MQCBDO_NONE, MQLONG_TYPE],
                ['CallbackArea', 0, 'P'],
                ['CallbackFunction', 0, 'P'],
                ['CallbackName', b'', '128s'],
                ['MaxMsgLength', CMQC.MQCBD_FULL_MSG_LENGTH, MQLONG_TYPE]]

        # Some padding is needed to match the C structure, even though it's not
        # listed in cmqc.h
        opts += [['_Reserved1', 0, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)
